#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import logging
import os
import sys
from collections import OrderedDict
from pprint import pformat

from ludwig.data.postprocessing import postprocess
from ludwig.data.preprocessing import preprocess_for_prediction
from ludwig.features.feature_registries import output_type_registry
from ludwig.globals import LUDWIG_VERSION, is_on_master, set_on_master
from ludwig.globals import TRAIN_SET_METADATA_FILE_NAME
from ludwig.models.model import load_model_and_definition
from ludwig.utils.data_utils import save_csv
from ludwig.utils.data_utils import save_json
from ludwig.utils.misc import get_from_registry
from ludwig.utils.print_utils import logging_level_registry, repr_ordered_dict
from ludwig.utils.print_utils import print_boxed
from ludwig.utils.print_utils import print_ludwig


def full_predict(
        model_path,
        data_csv=None,
        data_hdf5=None,
        dataset_type='generic',
        split='test',
        batch_size=128,
        skip_save_unprocessed_output=False,
        output_directory='results',
        only_predictions=False,
        gpus=None,
        gpu_fraction=1.0,
        use_horovod=False,
        debug=False,
        **kwargs
):
    # setup directories and file names
    experiment_dir_name = output_directory
    suffix = 0
    while os.path.exists(experiment_dir_name):
        experiment_dir_name = output_directory + '_' + str(suffix)
        suffix += 1

    if is_on_master():
        logging.info('Dataset type: {}'.format(dataset_type))
        logging.info('Dataset path: {}'.format(
            data_csv if data_csv is not None else data_hdf5))
        logging.info('Model path: {}'.format(model_path))
        logging.info('Output path: {}'.format(experiment_dir_name))
        logging.info('')

    train_set_metadata_json_fp = os.path.join(
        model_path,
        TRAIN_SET_METADATA_FILE_NAME
    )

    # preprocessing
    dataset, train_set_metadata = preprocess_for_prediction(
        model_path,
        split,
        dataset_type,
        data_csv,
        data_hdf5,
        train_set_metadata_json_fp,
        only_predictions
    )

    # run the prediction
    if is_on_master():
        print_boxed('LOADING MODEL')
    model, model_definition = load_model_and_definition(model_path,
                                                        use_horovod=use_horovod)

    prediction_results = predict(
        dataset,
        train_set_metadata,
        model,
        model_definition,
        batch_size,
        only_predictions,
        gpus,
        gpu_fraction,
        debug
    )
    model.close_session()

    if is_on_master():
        os.mkdir(experiment_dir_name)

        # postprocess
        postprocessed_output = postprocess(
            prediction_results,
            model_definition['output_features'],
            train_set_metadata,
            experiment_dir_name,
            skip_save_unprocessed_output or not is_on_master()
        )

        save_prediction_outputs(postprocessed_output, experiment_dir_name)

        if not only_predictions:
            print_prediction_results(prediction_results)
            save_prediction_statistics(prediction_results, experiment_dir_name)

        logging.info('Saved to: {0}'.format(experiment_dir_name))


def predict(
        dataset,
        train_set_metadata,
        model,
        model_definition,
        batch_size=128,
        only_predictions=False,
        gpus=None,
        gpu_fraction=1.0,
        debug=False
):
    """Computes predictions based on the computed model.
        :param dataset: Dataset contaning the data to calculate
               the predictions from.
        :type dataset: Dataset
        :param model: The trained model used to produce the predictions.
        :type model: Model
        :param model_definition: The model definition of the model to use
               for obtaining predictions
        :type model_definition: Dictionary
        :param batch_size: The size of batches when computing the predictions.
        :type batch_size: Integer
        :param only_predictions: If this parameter is True, only the predictions
               will be returned, if it is False, also performance metrics
               will be calculated on the predictions. It requires the data
               to contanin also ground truth for the output features, otherwise
               the metrics cannot be computed.
        :type only_predictions: Bool
        :type gpus: List
        :type gpu_fraction: Integer
        :param debug: If true turns on tfdbg with inf_or_nan checks.
        :type debug: Boolean

        :returns: A dictionary contaning the predictions of each output feature,
                  alongside with statistics on the quality of those predictions
                  (if only_predictions is False).
        """
    if is_on_master():
        print_boxed('PREDICT')
    test_stats = model.predict(
        dataset,
        batch_size,
        only_predictions=only_predictions,
        gpus=gpus,
        gpu_fraction=gpu_fraction
    )

    if not only_predictions:
        calculate_overall_stats(
            test_stats,
            model_definition['output_features'],
            dataset,
            train_set_metadata
        )

    return test_stats


def calculate_overall_stats(test_stats, output_features, dataset,
                            train_set_metadata):
    for output_feature in output_features:
        feature = get_from_registry(
            output_feature['type'],
            output_type_registry
        )
        feature.calculate_overall_stats(
            test_stats, output_feature, dataset, train_set_metadata
        )


def save_prediction_outputs(
        postprocessed_output,
        experiment_dir_name,
        skip_output_types=None
):
    if skip_output_types is None:
        skip_output_types = set()
    csv_filename = os.path.join(experiment_dir_name, '{}_{}.csv')
    for output_field, outputs in postprocessed_output.items():
        for output_type, values in outputs.items():
            if output_type not in skip_output_types:
                save_csv(csv_filename.format(output_field, output_type), values)


def save_prediction_statistics(prediction_stats, experiment_dir_name):
    test_stats_fn = os.path.join(
        experiment_dir_name,
        'prediction_statistics.json'
    )
    save_json(test_stats_fn, prediction_stats)


def print_prediction_results(prediction_stats):
    for output_field, result in prediction_stats.items():
        if (output_field != 'combined' or
                (output_field == 'combined' and len(prediction_stats) > 2)):
            logging.info('\n===== {} ====='.format(output_field))
            for measure in sorted(list(result)):
                if measure != 'confusion_matrix' and measure != 'roc_curve':
                    value = result[measure]
                    if isinstance(value, OrderedDict):
                        value_repr = repr_ordered_dict(value)
                    else:
                        value_repr = pformat(result[measure], indent=2)
                    logging.info(
                        '{0}: {1}'.format(
                            measure,
                            value_repr
                        )
                    )


def cli(sys_argv):
    parser = argparse.ArgumentParser(
        description='This script loads a pretrained model '
                    'and uses it to predict.',
        prog='ludwig predict',
        usage='%(prog)s [options]'
    )

    # ---------------
    # Data parameters
    # ---------------
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '--data_csv',
        help='input data CSV file. '
             'If it has a split column, it will be used for splitting '
             '(0: train, 1: validation, 2: test), '
             'otherwise the dataset will be randomly split'
    )
    group.add_argument(
        '--data_hdf5',
        help='input data HDF5 file. It is an intermediate preprocess version of'
             ' the input CSV created the first time a CSV file is used in the '
             'same directory with the same name and a hdf5 extension'
    )
    parser.add_argument(
        '--train_set_metadata_json',
        help='input metadata JSON file. It is an intermediate preprocess file '
             'containing the mappings of the input CSV created the first time '
             'a CSV file is used in the same directory with the same name and '
             'a json extension'
    )

    parser.add_argument(
        '-s',
        '--split',
        default='test',
        choices=['training', 'validation', 'test', 'full'],
        help='the split to test the model on'
    )

    # ----------------
    # Model parameters
    # ----------------
    parser.add_argument(
        '-m',
        '--model_path',
        help='model to load',
        required=True
    )

    # -------------------------
    # Output results parameters
    # -------------------------
    parser.add_argument(
        '-od',
        '--output_directory',
        type=str,
        default='results',
        help='directory that contains the results'
    )
    parser.add_argument(
        '-ssuo',
        '--skip_save_unprocessed_output',
        help='skips saving intermediate NPY output files',
        action='store_true', default=False
    )

    # ------------------
    # Generic parameters
    # ------------------
    parser.add_argument(
        '-bs',
        '--batch_size',
        type=int,
        default=128,
        help='size of batches'
    )
    parser.add_argument(
        '-op',
        '--only_predictions',
        action='store_true',
        default=False,
        help='skip metrics calculation'
    )

    # ------------------
    # Runtime parameters
    # ------------------
    parser.add_argument(
        '-g',
        '--gpus',
        type=int,
        default=0,
        help='list of gpu to use'
    )
    parser.add_argument(
        '-gf',
        '--gpu_fraction',
        type=float,
        default=1.0,
        help='fraction of gpu memory to initialize the process with'
    )
    parser.add_argument(
        '-uh',
        '--use_horovod',
        action='store_true',
        default=False,
        help='uses horovod for distributed training'
    )
    parser.add_argument(
        '-dbg',
        '--debug',
        action='store_true',
        default=False,
        help='enables debugging mode'
    )
    parser.add_argument(
        '-l',
        '--logging_level',
        default='info',
        help='the level of logging to use',
        choices=['critical', 'error', 'warning', 'info', 'debug', 'notset']
    )

    args = parser.parse_args(sys_argv)

    logging.basicConfig(
        stream=sys.stdout,
        level=logging_level_registry[args.logging_level],
        format='%(message)s'
    )

    set_on_master(args.use_horovod)

    if is_on_master():
        print_ludwig('Predict', LUDWIG_VERSION)

    full_predict(**vars(args))


if __name__ == '__main__':
    cli(sys.argv[1:])



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from collections import OrderedDict
from pprint import pformat

logging_level_registry = {
    'critical': logging.CRITICAL,
    'error': logging.ERROR,
    'warning': logging.WARNING,
    'info': logging.INFO,
    'debug': logging.DEBUG,
    'notset': logging.NOTSET
}


def print_ludwig(message, ludwig_version):
    logging.info('\n'.join([' _         _        _      ',
                            '| |_  _ __| |_ __ _(_)__ _ ',
                            '| | || / _` \ V  V / / _` |',
                            '|_|\_,_\__,_|\_/\_/|_\__, |',
                            '                     |___/ ',
                            'ludwig v{1} - {0}'.format(message, ludwig_version),
                            '']
                           ))


def print_boxed(text, print_fun=logging.info):
    box_width = len(text) + 2
    print_fun('')
    print_fun('╒{}╕'.format('═' * box_width))
    print_fun('│ {} │'.format(text.upper()))
    print_fun('╘{}╛'.format('═' * box_width))
    print_fun('')


def repr_ordered_dict(d: OrderedDict):
    return '{' + ',\n  '.join('{}: {}'.format(x, pformat(y, indent=4))
                              for x, y in d.items()) + '}'



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging

import tensorflow as tf

from ludwig.models.modules.convolutional_modules import ConvStack1D, \
    StackParallelConv1D, ParallelConv1D
from ludwig.models.modules.embedding_modules import EmbedSequence
from ludwig.models.modules.fully_connected_modules import FCStack
from ludwig.models.modules.recurrent_modules import RecurrentStack
from ludwig.models.modules.recurrent_modules import reduce_sequence


class EmbedEncoder:

    def __init__(
            self,
            vocab,
            representation='dense',
            embedding_size=256,
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=False,
            dropout=False,
            initializer=None,
            regularize=True,
            reduce_output='sum',
            **kwargs
    ):
        """
            :param should_embed: If True the input sequence is expected
                   to be made of integers and will be mapped into embeddings
            :type should_embed: Boolean
            :param vocab: Vocabulary of the input feature to encode
            :type vocab: List
            :param representation: the possible values are `dense` and `sparse`.
                   `dense` means the mebeddings are initialized randomly,
                   `sparse` meanse they are initialized to be one-hot encodings.
            :type representation: Str (one of 'dense' or 'sparse')
            :param embedding_size: it is the maximum embedding size, the actual
                   size will be `min(vocaularyb_size, embedding_size)`
                   for `dense` representations and exacly `vocaularyb_size`
                   for the `sparse` encoding, where `vocabulary_size` is
                   the number of different strings appearing in the training set
                   in the column the feature is named after (plus 1 for `<UNK>`).
            :type embedding_size: Integer
            :param embeddings_trainable: If `True` embeddings are trained during
                   the training process, if `False` embeddings are fixed.
                   It may be useful when loading pretrained embeddings
                   for avoiding finetuning them. This parameter has effect only
                   for `representation` is `dense` as `sparse` one-hot encodings
                    are not trainable.
            :type embeddings_trainable: Boolean
            :param pretrained_embeddings: by default `dense` embeddings
                   are initialized randomly, but this parameter allow to specify
                   a path to a file containing embeddings in the GloVe format.
                   When the file containing the embeddings is loaded, only the
                   embeddings with labels present in the vocabulary are kept,
                   the others are discarded. If the vocabulary contains strings
                   that have no match in the embeddings file, their embeddings
                   are initialized with the average of all other embedding plus
                   some random noise to make them different from each other.
                   This parameter has effect only if `representation` is `dense`.
            :type pretrained_embeddings: str (filepath)
            :param embeddings_on_cpu: by default embedings matrices are stored
                   on GPU memory if a GPU is used, as it allows
                   for faster access, but in some cases the embedding matrix
                   may be really big and this parameter forces the placement
                   of the embedding matrix in regular memroy and the CPU is used
                   to resolve them, slightly slowing down the process
                   as a result of data transfer between CPU and GPU memory.
            :param dropout: determines if there should be a dropout layer before
                   returning the encoder output.
            :type dropout: Boolean
            :param initializer: the initializer to use. If `None`, the default
                   initialized of each variable is used (`glorot_uniform`
                   in most cases). Options are: `constant`, `identity`, `zeros`,
                    `ones`, `orthogonal`, `normal`, `uniform`,
                    `truncated_normal`, `variance_scaling`, `glorot_normal`,
                    `glorot_uniform`, `xavier_normal`, `xavier_uniform`,
                    `he_normal`, `he_uniform`, `lecun_normal`, `lecun_uniform`.
                    Alternatively it is possible to specify a dictionary with
                    a key `type` that identifies the type of initialzier and
                    other keys for its parameters, e.g.
                    `{type: normal, mean: 0, stddev: 0}`.
                    To know the parameters of each initializer, please refer to
                    TensorFlow's documentation.
            :type initializer: str
            :param regularize: if `True` the embedding wieghts are added to
                   the set of weights that get reularized by a regularization
                   loss (if the `regularization_lambda` in `training`
                   is greater than 0).
            :type regularize: Boolean
            :param reduce_output: defines how to reduce the output tensor along
                   the `s` sequence length dimention if the rank of the tensor
                   is greater than 2. Available values are: `sum`,
                   `mean` or `avg`, `max`, `concat` (concatenates along
                   the first dimension), `last` (returns the last vector of the
                   first dimension) and `None` or `null` (which does not reduce
                   and returns the full tensor).
            :type reduce_output: str
        """
        self.reduce_output = reduce_output

        self.embed_sequence = EmbedSequence(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            initializer=initializer,
            regularize=regularize
        )

    def __call__(
            self,
            input_sequence,
            regularizer,
            dropout_rate,
            is_training=True
    ):
        """
            :param input_sequence: The input sequence fed into the encoder.
                   Shape: [batch x sequence length], type tf.int32
            :type input_sequence: Tensor
            :param regularizer: The regularizer to use for the weights
                   of the encoder.
            :type regularizer:
            :param dropout_rate: Tensor (tf.float) of the probability of dropout
            :type dropout_rate: Tensor
            :param is_training: Tesnor (tf.bool) specifying if in training mode
                   (important for dropout)
            :type is_training: Tensor
        """
        # ================ Embeddings ================
        embedded_sequence, embedding_size = self.embed_sequence(
            input_sequence,
            regularizer,
            dropout_rate,
            is_training=True
        )

        hidden = reduce_sequence(embedded_sequence, self.reduce_output)

        return hidden, embedding_size


class ParallelCNN(object):

    def __init__(
            self,
            should_embed=True,
            vocab=None,
            representation='dense',
            embedding_size=256,
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=False,
            conv_layers=None,
            num_conv_layers=None,
            filter_size=3,
            num_filters=256,
            pool_size=None,
            fc_layers=None,
            num_fc_layers=None,
            fc_size=256,
            norm=None,
            activation='relu',
            dropout=False,
            initializer=None,
            regularize=True,
            reduce_output='max',
            **kwargs):
        """
            :param should_embed: If True the input sequence is expected
                   to be made of integers and will be mapped into embeddings
            :type should_embed: Boolean
            :param vocab: Vocabulary of the input feature to encode
            :type vocab: List
            :param representation: the possible values are `dense` and `sparse`.
                   `dense` means the mebeddings are initialized randomly,
                   `sparse` meanse they are initialized to be one-hot encodings.
            :type representation: Str (one of 'dense' or 'sparse')
            :param embedding_size: it is the maximum embedding size, the actual
                   size will be `min(vocaularyb_size, embedding_size)`
                   for `dense` representations and exacly `vocaularyb_size`
                   for the `sparse` encoding, where `vocabulary_size` is
                   the number of different strings appearing in the training set
                   in the column the feature is named after (plus 1 for `<UNK>`).
            :type embedding_size: Integer
            :param embeddings_trainable: If `True` embeddings are trained during
                   the training process, if `False` embeddings are fixed.
                   It may be useful when loading pretrained embeddings
                   for avoiding finetuning them. This parameter has effect only
                   for `representation` is `dense` as `sparse` one-hot encodings
                    are not trainable.
            :type embeddings_trainable: Boolean
            :param pretrained_embeddings: by default `dense` embeddings
                   are initialized randomly, but this parameter allow to specify
                   a path to a file containing embeddings in the GloVe format.
                   When the file containing the embeddings is loaded, only the
                   embeddings with labels present in the vocabulary are kept,
                   the others are discarded. If the vocabulary contains strings
                   that have no match in the embeddings file, their embeddings
                   are initialized with the average of all other embedding plus
                   some random noise to make them different from each other.
                   This parameter has effect only if `representation` is `dense`.
            :type pretrained_embeddings: str (filepath)
            :param embeddings_on_cpu: by default embedings matrices are stored
                   on GPU memory if a GPU is used, as it allows
                   for faster access, but in some cases the embedding matrix
                   may be really big and this parameter forces the placement
                   of the embedding matrix in regular memroy and the CPU is used
                   to resolve them, slightly slowing down the process
                   as a result of data transfer between CPU and GPU memory.
            :param conv_layers: it is a list of dictionaries containing
                   the parameters of all the convolutional layers. The length
                   of the list determines the number of parallel convolutional
                   layers and the content of each dictionary determines
                   the parameters for a specific layer. The available parameters
                   for each layer are: `filter_size`, `num_filters`, `pool`,
                   `norm`, `activation` and `regularize`. If any of those values
                   is missing from the dictionary, the default one specified
                   as a parameter of the encoder will be used instead. If both
                   `conv_layers` and `num_conv_layers` are `None`, a default
                   list will be assigned to `conv_layers` with the value
                   `[{filter_size: 2}, {filter_size: 3}, {filter_size: 4},
                   {filter_size: 5}]`.
            :type conv_layers: List
            :param num_conv_layers: if `conv_layers` is `None`, this is
                   the number of parallel convolutional layers.
            :type num_conv_layers: Integer
            :param filter_size:  if a `filter_size` is not already specified in
                   `conv_layers` this is the default `filter_size` that
                   will be used for each layer. It indicates how wide is
                   the 1d convolutional filter.
            :type filter_size: Integer
            :param num_filters: if a `num_filters` is not already specified in
                   `conv_layers` this is the default `num_filters` that
                   will be used for each layer. It indicates the number
                   of filters, and by consequence the output channels of
                   the 1d convolution.
            :type num_filters: Integer
            :param pool_size: if a `pool_size` is not already specified
                  in `conv_layers` this is the default `pool_size` that
                  will be used for each layer. It indicates the size of
                  the max pooling that will be performed along the `s` sequence
                  dimension after the convolution operation.
            :type pool_size: Integer
            :param fc_layers: it is a list of dictionaries containing
                   the parameters of all the fully connected layers. The length
                   of the list determines the number of stacked fully connected
                   layers and the content of each dictionary determines
                   the parameters for a specific layer. The available parameters
                   for each layer are: `fc_size`, `norm`, `activation` and
                   `regularize`. If any of those values is missing from
                   the dictionary, the default one specified as a parameter of
                   the encoder will be used instead. If both `fc_layers` and
                   `num_fc_layers` are `None`, a default list will be assigned
                   to `fc_layers` with the value
                   `[{fc_size: 512}, {fc_size: 256}]`.
                   (only applies if `reduce_output` is not `None`).
            :type fc_layers: List
            :param num_fc_layers: if `fc_layers` is `None`, this is the number
                   of stacked fully connected layers (only applies if
                   `reduce_output` is not `None`).
            :type num_fc_layers: Integer
            :param fc_size: if a `fc_size` is not already specified in
                   `fc_layers` this is the default `fc_size` that will be used
                   for each layer. It indicates the size of the output
                   of a fully connected layer.
            :type fc_size: Integer
            :param norm: if a `norm` is not already specified in `conv_layers`
                   or `fc_layers` this is the default `norm` that will be used
                   for each layer. It indicates the norm of the output.
            :type norm: str
            :param activation: Default activation function to use
            :type activation: Str
            :param dropout: determines if there should be a dropout layer before
                   returning the encoder output.
            :type dropout: Boolean
            :param initializer: the initializer to use. If `None` it uses
                   `glorot_uniform`. Options are: `constant`, `identity`,
                   `zeros`, `ones`, `orthogonal`, `normal`, `uniform`,
                   `truncated_normal`, `variance_scaling`, `glorot_normal`,
                   `glorot_uniform`, `xavier_normal`, `xavier_uniform`,
                   `he_normal`, `he_uniform`, `lecun_normal`, `lecun_uniform`.
                   Alternatively it is possible to specify a dictionary with
                   a key `type` that identifies the type of initialzier and
                   other keys for its parameters,
                   e.g. `{type: normal, mean: 0, stddev: 0}`.
                   To know the parameters of each initializer, please refer
                   to TensorFlow's documentation.
            :type initializer: str
            :param regularize: if a `regularize` is not already specified in
                   `conv_layers` or `fc_layers` this is the default `regularize`
                   that will be used for each layer. It indicates if
                   the layer weights should be considered when computing
                   a regularization loss.
            :type regularize:
            :param reduce_output: defines how to reduce the output tensor of
                   the convolutional layers along the `s` sequence length
                   dimention if the rank of the tensor is greater than 2.
                   Available values are: `sum`, `mean` or `avg`, `max`, `concat`
                   (concatenates along the first dimension), `last` (returns
                   the last vector of the first dimension) and `None` or `null`
                   (which does not reduce and returns the full tensor).
            :type reduce_output: str
        """
        self.should_embed = should_embed

        if conv_layers is not None and num_conv_layers is None:
            # use custom-defined layers
            self.conv_layers = conv_layers
            self.num_conv_layers = len(conv_layers)
        elif conv_layers is None and num_conv_layers is not None:
            # generate num_conv_layers with default parameters
            self.conv_layers = None
            self.num_conv_layers = num_conv_layers
        elif conv_layers is None and num_conv_layers is None:
            # use default layers with varying filter sizes
            self.conv_layers = [
                {'filter_size': 2},
                {'filter_size': 3},
                {'filter_size': 4},
                {'filter_size': 5}
            ]
            self.num_conv_layers = 4
        else:
            raise ValueError(
                'Invalid layer parametrization, use either conv_layers or num_conv_layers')

        if fc_layers is not None and num_fc_layers is None:
            # use custom-defined layers
            fc_layers = fc_layers
            num_fc_layers = len(fc_layers)
        elif fc_layers is None and num_fc_layers is not None:
            # generate num_fc_layers with default parameters
            fc_layers = None
            num_fc_layers = num_fc_layers
        elif fc_layers is None and num_fc_layers is None:
            # use default layers with varying filter sizes
            fc_layers = [
                {'fc_size': 512},
                {'fc_size': 256}
            ]
            num_fc_layers = 2
        else:
            raise ValueError(
                'Invalid layer parametrization, use either fc_layers or num_fc_layers')

        self.reduce_output = reduce_output

        self.embed_sequence = EmbedSequence(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            initializer=initializer,
            regularize=regularize
        )

        self.parallel_conv_1d = ParallelConv1D(
            layers=self.conv_layers,
            default_filter_size=filter_size,
            default_num_filters=num_filters,
            default_pool_size=pool_size,
            default_activation=activation,
            default_norm=norm,
            default_dropout=dropout,
            default_initializer=initializer,
            default_regularize=regularize
        )

        self.fc_stack = FCStack(
            layers=fc_layers,
            num_layers=num_fc_layers,
            default_fc_size=fc_size,
            default_activation=activation,
            default_norm=norm,
            default_dropout=dropout,
            default_regularize=regularize,
            default_initializer=initializer
        )

    def __call__(
            self,
            input_sequence,
            regularizer,
            dropout_rate,
            is_training=True
    ):
        """
            :param input_sequence: The input sequence fed into the encoder.
                   Shape: [batch x sequence length], type tf.int32
            :type input_sequence: Tensor
            :param regularizer: The regularizer to use for the weights
                   of the encoder.
            :type regularizer:
            :param dropout_rate: Tensor (tf.float) of the probability of dropout
            :type dropout_rate: Tensor
            :param is_training: Tesnor (tf.bool) specifying if in training mode
                   (important for dropout)
            :type is_training: Tensor
        """
        # ================ Embeddings ================
        if self.should_embed:
            embedded_input_sequence, embedding_size = self.embed_sequence(
                input_sequence,
                regularizer,
                dropout_rate,
                is_training=True
            )
        else:
            embedded_input_sequence = input_sequence
            while len(embedded_input_sequence.shape) < 3:
                embedded_input_sequence = tf.expand_dims(
                    embedded_input_sequence, -1)
            embedding_size = 1

        # shape=(?, sequence_length, embedding_size)
        hidden = embedded_input_sequence
        logging.debug('  hidden: {0}'.format(hidden))

        # ================ Conv Layers ================
        hidden = self.parallel_conv_1d(
            hidden,
            embedding_size,
            regularizer=regularizer,
            dropout_rate=dropout_rate,
            is_training=is_training
        )
        hidden_size = sum(
            [conv_layer['num_filters'] for conv_layer in self.conv_layers]
        )
        logging.debug('  hidden: {0}'.format(hidden))

        # ================ Sequence Reduction ================
        if self.reduce_output is not None:
            hidden = reduce_sequence(hidden, self.reduce_output)

            # ================ FC Layers ================
            hidden_size = hidden.shape.as_list()[-1]
            logging.debug('  flatten hidden: {0}'.format(hidden))

            hidden = self.fc_stack(
                hidden,
                hidden_size,
                regularizer=regularizer,
                dropout_rate=dropout_rate,
                is_training=is_training
            )
            hidden_size = hidden.shape.as_list()[-1]

        return hidden, hidden_size


class StackedCNN:

    def __init__(
            self,
            should_embed=True,
            vocab=None,
            representation='dense',
            embedding_size=256,
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=False,
            conv_layers=None,
            num_conv_layers=None,
            filter_size=5,
            num_filters=256,
            pool_size=None,
            fc_layers=None,
            num_fc_layers=None,
            fc_size=256,
            norm=None,
            activation='relu',
            dropout=False,
            initializer=None,
            regularize=True,
            reduce_output='max',
            **kwargs
    ):
        """
            :param should_embed: If True the input sequence is expected
                   to be made of integers and will be mapped into embeddings
            :type should_embed: Boolean
            :param vocab: Vocabulary of the input feature to encode
            :type vocab: List
            :param representation: the possible values are `dense` and `sparse`.
                   `dense` means the mebeddings are initialized randomly,
                   `sparse` meanse they are initialized to be one-hot encodings.
            :type representation: Str (one of 'dense' or 'sparse')
            :param embedding_size: it is the maximum embedding size, the actual
                   size will be `min(vocaularyb_size, embedding_size)`
                   for `dense` representations and exacly `vocaularyb_size`
                   for the `sparse` encoding, where `vocabulary_size` is
                   the number of different strings appearing in the training set
                   in the column the feature is named after (plus 1 for `<UNK>`).
            :type embedding_size: Integer
            :param embeddings_trainable: If `True` embeddings are trained during
                   the training process, if `False` embeddings are fixed.
                   It may be useful when loading pretrained embeddings
                   for avoiding finetuning them. This parameter has effect only
                   for `representation` is `dense` as `sparse` one-hot encodings
                    are not trainable.
            :type embeddings_trainable: Boolean
            :param pretrained_embeddings: by default `dense` embeddings
                   are initialized randomly, but this parameter allow to specify
                   a path to a file containing embeddings in the GloVe format.
                   When the file containing the embeddings is loaded, only the
                   embeddings with labels present in the vocabulary are kept,
                   the others are discarded. If the vocabulary contains strings
                   that have no match in the embeddings file, their embeddings
                   are initialized with the average of all other embedding plus
                   some random noise to make them different from each other.
                   This parameter has effect only if `representation` is `dense`.
            :type pretrained_embeddings: str (filepath)
            :param embeddings_on_cpu: by default embedings matrices are stored
                   on GPU memory if a GPU is used, as it allows
                   for faster access, but in some cases the embedding matrix
                   may be really big and this parameter forces the placement
                   of the embedding matrix in regular memroy and the CPU is used
                   to resolve them, slightly slowing down the process
                   as a result of data transfer between CPU and GPU memory.
            :param conv_layers: it is a list of dictionaries containing
                   the parameters of all the convolutional layers. The length
                   of the list determines the number of parallel convolutional
                   layers and the content of each dictionary determines
                   the parameters for a specific layer. The available parameters
                   for each layer are: `filter_size`, `num_filters`, `pool`,
                   `norm`, `activation` and `regularize`. If any of those values
                   is missing from the dictionary, the default one specified
                   as a parameter of the encoder will be used instead. If both
                   `conv_layers` and `num_conv_layers` are `None`, a default
                   list will be assigned to `conv_layers` with the value
                   `[{filter_size: 2}, {filter_size: 3}, {filter_size: 4},
                   {filter_size: 5}]`.
            :type conv_layers: List
            :param num_conv_layers: if `conv_layers` is `None`, this is
                   the number of stacked convolutional layers.
            :type num_conv_layers: Integer
            :param filter_size:  if a `filter_size` is not already specified in
                   `conv_layers` this is the default `filter_size` that
                   will be used for each layer. It indicates how wide is
                   the 1d convolutional filter.
            :type filter_size: Integer
            :param num_filters: if a `num_filters` is not already specified in
                   `conv_layers` this is the default `num_filters` that
                   will be used for each layer. It indicates the number
                   of filters, and by consequence the output channels of
                   the 1d convolution.
            :type num_filters: Integer
            :param pool_size: if a `pool_size` is not already specified
                  in `conv_layers` this is the default `pool_size` that
                  will be used for each layer. It indicates the size of
                  the max pooling that will be performed along the `s` sequence
                  dimension after the convolution operation.
            :type pool_size: Integer
            :param fc_layers: it is a list of dictionaries containing
                   the parameters of all the fully connected layers. The length
                   of the list determines the number of stacked fully connected
                   layers and the content of each dictionary determines
                   the parameters for a specific layer. The available parameters
                   for each layer are: `fc_size`, `norm`, `activation` and
                   `regularize`. If any of those values is missing from
                   the dictionary, the default one specified as a parameter of
                   the encoder will be used instead. If both `fc_layers` and
                   `num_fc_layers` are `None`, a default list will be assigned
                   to `fc_layers` with the value
                   `[{fc_size: 512}, {fc_size: 256}]`.
                   (only applies if `reduce_output` is not `None`).
            :type fc_layers: List
            :param num_fc_layers: if `fc_layers` is `None`, this is the number
                   of stacked fully connected layers (only applies if
                   `reduce_output` is not `None`).
            :type num_fc_layers: Integer
            :param fc_size: if a `fc_size` is not already specified in
                   `fc_layers` this is the default `fc_size` that will be used
                   for each layer. It indicates the size of the output
                   of a fully connected layer.
            :type fc_size: Integer
            :param norm: if a `norm` is not already specified in `conv_layers`
                   or `fc_layers` this is the default `norm` that will be used
                   for each layer. It indicates the norm of the output.
            :type norm: str
            :param activation: Default activation function to use
            :type activation: Str
            :param dropout: determines if there should be a dropout layer before
                   returning the encoder output.
            :type dropout: Boolean
            :param initializer: the initializer to use. If `None` it uses
                   `glorot_uniform`. Options are: `constant`, `identity`,
                   `zeros`, `ones`, `orthogonal`, `normal`, `uniform`,
                   `truncated_normal`, `variance_scaling`, `glorot_normal`,
                   `glorot_uniform`, `xavier_normal`, `xavier_uniform`,
                   `he_normal`, `he_uniform`, `lecun_normal`, `lecun_uniform`.
                   Alternatively it is possible to specify a dictionary with
                   a key `type` that identifies the type of initialzier and
                   other keys for its parameters,
                   e.g. `{type: normal, mean: 0, stddev: 0}`.
                   To know the parameters of each initializer, please refer
                   to TensorFlow's documentation.
            :type initializer: str
            :param regularize: if a `regularize` is not already specified in
                   `conv_layers` or `fc_layers` this is the default `regularize`
                   that will be used for each layer. It indicates if
                   the layer weights should be considered when computing
                   a regularization loss.
            :type regularize:
            :param reduce_output: defines how to reduce the output tensor of
                   the convolutional layers along the `s` sequence length
                   dimention if the rank of the tensor is greater than 2.
                   Available values are: `sum`, `mean` or `avg`, `max`, `concat`
                   (concatenates along the first dimension), `last` (returns
                   the last vector of the first dimension) and `None` or `null`
                   (which does not reduce and returns the full tensor).
            :type reduce_output: str
        """
        if conv_layers is not None and num_conv_layers is None:
            # use custom-defined layers
            self.conv_layers = conv_layers
            self.num_conv_layers = len(conv_layers)
        elif conv_layers is None and num_conv_layers is not None:
            # generate num_conv_layers with default parameters
            self.conv_layers = None
            self.num_conv_layers = num_conv_layers
        elif conv_layers is None and num_conv_layers is None:
            # use default layers with varying filter sizes
            self.conv_layers = [
                {
                    'filter_size': 7,
                    'pool_size': 3,
                    'regularize': False
                },
                {
                    'filter_size': 7,
                    'pool_size': 3,
                    'regularize': False
                },
                {
                    'filter_size': 3,
                    'pool_size': None,
                    'regularize': False
                },
                {
                    'filter_size': 3,
                    'pool_size': None,
                    'regularize': False
                },
                {
                    'filter_size': 3,
                    'pool_size': None,
                    'regularize': True
                },
                {
                    'filter_size': 3,
                    'pool_size': 3,
                    'regularize': True
                }
            ]
            self.num_conv_layers = 6
        else:
            raise ValueError(
                'Invalid layer parametrization, use either conv_layers or num_conv_layers')

        if fc_layers is not None and num_fc_layers is None:
            # use custom-defined layers
            fc_layers = fc_layers
            num_fc_layers = len(fc_layers)
        elif fc_layers is None and num_fc_layers is not None:
            # generate num_fc_layers with default parameters
            fc_layers = None
            num_fc_layers = num_fc_layers
        elif fc_layers is None and num_fc_layers is None:
            # use default layers with varying filter sizes
            fc_layers = [
                {'fc_size': 512},
                {'fc_size': 256}
            ]
            num_fc_layers = 2
        else:
            raise ValueError(
                'Invalid layer parametrization, use either fc_layers or num_fc_layers')

        self.should_embed = should_embed
        self.reduce_output = reduce_output

        self.embed_sequence = EmbedSequence(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            initializer=initializer,
            regularize=regularize
        )

        self.conv_stack_1d = ConvStack1D(
            layers=self.conv_layers,
            default_filter_size=filter_size,
            default_num_filters=num_filters,
            default_activation=activation,
            default_norm=norm,
            default_pool_size=pool_size,
            default_dropout=dropout,
            default_initializer=initializer,
            default_regularize=regularize
        )

        self.fc_stack = FCStack(
            layers=fc_layers,
            num_layers=num_fc_layers,
            default_fc_size=fc_size,
            default_activation=activation,
            default_norm=norm,
            default_dropout=dropout,
            default_regularize=regularize,
            default_initializer=initializer
        )

    def __call__(
            self,
            input_sequence,
            regularizer,
            dropout_rate,
            is_training=True
    ):
        """
            :param input_sequence: The input sequence fed into the encoder.
                   Shape: [batch x sequence length], type tf.int32
            :type input_sequence: Tensor
            :param regularizer: The regularizer to use for the weights
                   of the encoder.
            :type regularizer:
            :param dropout_rate: Tensor (tf.float) of the probability of dropout
            :type dropout_rate: Tensor
            :param is_training: Tesnor (tf.bool) specifying if in training mode
                   (important for dropout)
            :type is_training: Tensor
        """
        # ================ Embeddings ================
        if self.should_embed:
            embedded_input_sequence, self.embedding_size = self.embed_sequence(
                input_sequence,
                regularizer,
                dropout_rate,
                is_training=True
            )
        else:
            embedded_input_sequence = input_sequence
            while len(embedded_input_sequence.shape) < 3:
                embedded_input_sequence = tf.expand_dims(
                    embedded_input_sequence, -1)
            self.embedding_size = embedded_input_sequence.shape[-1]

        hidden = embedded_input_sequence
        logging.debug('  hidden: {0}'.format(hidden))

        # ================ Conv Layers ================
        with tf.variable_scope('stack_conv'):
            hidden = self.conv_stack_1d(
                hidden,
                self.embedding_size,
                regularizer=regularizer,
                dropout_rate=dropout_rate,
                is_training=is_training
            )
        hidden_size = self.conv_layers[-1]['num_filters']
        logging.debug('  hidden: {0}'.format(hidden))

        # ================ Sequence Reduction ================
        if self.reduce_output is not None:
            hidden = reduce_sequence(hidden, self.reduce_output)

            # ================ FC Layers ================
            hidden_size = hidden.shape.as_list()[-1]
            logging.debug('  flatten hidden: {0}'.format(hidden))

            hidden = self.fc_stack(
                hidden,
                hidden_size,
                regularizer=regularizer,
                dropout_rate=dropout_rate,
                is_training=is_training
            )
            hidden_size = hidden.shape.as_list()[-1]

        return hidden, hidden_size


class StackedParallelCNN:

    def __init__(
            self,
            should_embed=True,
            vocab=None,
            representation='dense',
            embedding_size=256,
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=False,
            stacked_layers=None,
            num_stacked_layers=None,
            filter_size=3,
            num_filters=256,
            stride=1,
            pool_size=None,
            pool_stride=1,
            fc_layers=None,
            num_fc_layers=None,
            fc_size=256,
            norm=None,
            activation='relu',
            dropout=False,
            initializer=None,
            regularize=True,
            reduce_output='max',
            **kwargs
    ):
        """
            :param should_embed: If True the input sequence is expected
                   to be made of integers and will be mapped into embeddings
            :type should_embed: Boolean
            :param vocab: Vocabulary of the input feature to encode
            :type vocab: List
            :param representation: the possible values are `dense` and `sparse`.
                   `dense` means the mebeddings are initialized randomly,
                   `sparse` meanse they are initialized to be one-hot encodings.
            :type representation: Str (one of 'dense' or 'sparse')
            :param embedding_size: it is the maximum embedding size, the actual
                   size will be `min(vocaularyb_size, embedding_size)`
                   for `dense` representations and exacly `vocaularyb_size`
                   for the `sparse` encoding, where `vocabulary_size` is
                   the number of different strings appearing in the training set
                   in the column the feature is named after (plus 1 for `<UNK>`).
            :type embedding_size: Integer
            :param embeddings_trainable: If `True` embeddings are trained during
                   the training process, if `False` embeddings are fixed.
                   It may be useful when loading pretrained embeddings
                   for avoiding finetuning them. This parameter has effect only
                   for `representation` is `dense` as `sparse` one-hot encodings
                    are not trainable.
            :type embeddings_trainable: Boolean
            :param pretrained_embeddings: by default `dense` embeddings
                   are initialized randomly, but this parameter allow to specify
                   a path to a file containing embeddings in the GloVe format.
                   When the file containing the embeddings is loaded, only the
                   embeddings with labels present in the vocabulary are kept,
                   the others are discarded. If the vocabulary contains strings
                   that have no match in the embeddings file, their embeddings
                   are initialized with the average of all other embedding plus
                   some random noise to make them different from each other.
                   This parameter has effect only if `representation` is `dense`.
            :type pretrained_embeddings: str (filepath)
            :param embeddings_on_cpu: by default embedings matrices are stored
                   on GPU memory if a GPU is used, as it allows
                   for faster access, but in some cases the embedding matrix
                   may be really big and this parameter forces the placement
                   of the embedding matrix in regular memroy and the CPU is used
                   to resolve them, slightly slowing down the process
                   as a result of data transfer between CPU and GPU memory.
            :param stacked_layers: it is a of lists of list of dictionaries
                   containing the parameters of the stack of
                   parallel convolutional layers. The length of the list
                   determines the number of stacked parallel
                   convolutional layers, length of the sub-lists determines
                   the number of parallel conv layers and the content
                   of each dictionary determines the parameters for
                   a specific layer. The available parameters for each layer are:
                   `filter_size`, `num_filters`, `pool_size`, `norm`,
                   `activation` and `regularize`. If any of those values
                   is missing from the dictionary, the default one specified
                   as a parameter of the encoder will be used instead. If both
                   `stacked_layers` and `num_stacked_layers` are `None`,
                   a default list will be assigned to `stacked_layers` with
                   the value `[[{filter_size: 2}, {filter_size: 3},
                   {filter_size: 4}, {filter_size: 5}], [{filter_size: 2},
                   {filter_size: 3}, {filter_size: 4}, {filter_size: 5}],
                   [{filter_size: 2}, {filter_size: 3}, {filter_size: 4},
                   {filter_size: 5}]]`.
            :type stacked_layers: List
            :param num_stacked_layers: if `stacked_layers` is `None`, this is
                   the number of elements in the stack of
                   parallel convolutional layers.
            :type num_stacked_layers: Integer
            :param filter_size:  if a `filter_size` is not already specified in
                   `conv_layers` this is the default `filter_size` that
                   will be used for each layer. It indicates how wide is
                   the 1d convolutional filter.
            :type filter_size: Integer
            :param num_filters: if a `num_filters` is not already specified in
                   `conv_layers` this is the default `num_filters` that
                   will be used for each layer. It indicates the number
                   of filters, and by consequence the output channels of
                   the 1d convolution.
            :type num_filters: Integer
            :param pool_size: if a `pool_size` is not already specified
                  in `conv_layers` this is the default `pool_size` that
                  will be used for each layer. It indicates the size of
                  the max pooling that will be performed along the `s` sequence
                  dimension after the convolution operation.
            :type pool_size: Integer
            :param fc_layers: it is a list of dictionaries containing
                   the parameters of all the fully connected layers. The length
                   of the list determines the number of stacked fully connected
                   layers and the content of each dictionary determines
                   the parameters for a specific layer. The available parameters
                   for each layer are: `fc_size`, `norm`, `activation` and
                   `regularize`. If any of those values is missing from
                   the dictionary, the default one specified as a parameter of
                   the encoder will be used instead. If both `fc_layers` and
                   `num_fc_layers` are `None`, a default list will be assigned
                   to `fc_layers` with the value
                   `[{fc_size: 512}, {fc_size: 256}]`.
                   (only applies if `reduce_output` is not `None`).
            :type fc_layers: List
            :param num_fc_layers: if `fc_layers` is `None`, this is the number
                   of stacked fully connected layers (only applies if
                   `reduce_output` is not `None`).
            :type num_fc_layers: Integer
            :param fc_size: if a `fc_size` is not already specified in
                   `fc_layers` this is the default `fc_size` that will be used
                   for each layer. It indicates the size of the output
                   of a fully connected layer.
            :type fc_size: Integer
            :param norm: if a `norm` is not already specified in `conv_layers`
                   or `fc_layers` this is the default `norm` that will be used
                   for each layer. It indicates the norm of the output.
            :type norm: str
            :param activation: Default activation function to use
            :type activation: Str
            :param dropout: determines if there should be a dropout layer before
                   returning the encoder output.
            :type dropout: Boolean
            :param initializer: the initializer to use. If `None` it uses
                   `glorot_uniform`. Options are: `constant`, `identity`,
                   `zeros`, `ones`, `orthogonal`, `normal`, `uniform`,
                   `truncated_normal`, `variance_scaling`, `glorot_normal`,
                   `glorot_uniform`, `xavier_normal`, `xavier_uniform`,
                   `he_normal`, `he_uniform`, `lecun_normal`, `lecun_uniform`.
                   Alternatively it is possible to specify a dictionary with
                   a key `type` that identifies the type of initialzier and
                   other keys for its parameters,
                   e.g. `{type: normal, mean: 0, stddev: 0}`.
                   To know the parameters of each initializer, please refer
                   to TensorFlow's documentation.
            :type initializer: str
            :param regularize: if a `regularize` is not already specified in
                   `conv_layers` or `fc_layers` this is the default `regularize`
                   that will be used for each layer. It indicates if
                   the layer weights should be considered when computing
                   a regularization loss.
            :type regularize:
            :param reduce_output: defines how to reduce the output tensor of
                   the convolutional layers along the `s` sequence length
                   dimention if the rank of the tensor is greater than 2.
                   Available values are: `sum`, `mean` or `avg`, `max`, `concat`
                   (concatenates along the first dimension), `last` (returns
                   the last vector of the first dimension) and `None` or `null`
                   (which does not reduce and returns the full tensor).
            :type reduce_output: str
        """
        if stacked_layers is not None and num_stacked_layers is None:
            # use custom-defined layers
            self.stacked_layers = stacked_layers
            self.num_stacked_layers = len(stacked_layers)
        elif stacked_layers is None and num_stacked_layers is not None:
            # generate num_conv_layers with default parameters
            self.stacked_layers = None
            self.num_stacked_layers = num_stacked_layers
        elif stacked_layers is None and num_stacked_layers is None:
            # use default layers with varying filter sizes
            self.stacked_layers = [
                [
                    {'filter_size': 2},
                    {'filter_size': 3},
                    {'filter_size': 4},
                    {'filter_size': 5}
                ],
                [
                    {'filter_size': 2},
                    {'filter_size': 3},
                    {'filter_size': 4},
                    {'filter_size': 5}
                ],
                [
                    {'filter_size': 2},
                    {'filter_size': 3},
                    {'filter_size': 4},
                    {'filter_size': 5}
                ]
            ]
            self.num_stacked_layers = 6
        else:
            raise ValueError(
                'Invalid layer parametrization, use either stacked_layers or num_stacked_layers')

        if fc_layers is not None and num_fc_layers is None:
            # use custom-defined layers
            fc_layers = fc_layers
            num_fc_layers = len(fc_layers)
        elif fc_layers is None and num_fc_layers is not None:
            # generate num_fc_layers with default parameters
            fc_layers = None
            num_fc_layers = num_fc_layers
        elif fc_layers is None and num_fc_layers is None:
            # use default layers with varying filter sizes
            fc_layers = [
                {'fc_size': 512},
                {'fc_size': 256}
            ]
            num_fc_layers = 2
        else:
            raise ValueError(
                'Invalid layer parametrization, use either fc_layers or num_fc_layers')

        self.should_embed = should_embed
        self.reduce_output = reduce_output

        self.embed_sequence = EmbedSequence(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            initializer=initializer,
            regularize=regularize
        )

        self.stack_parallel_conv_1d = StackParallelConv1D(
            stacked_layers=self.stacked_layers,
            default_filter_size=filter_size,
            default_num_filters=num_filters,
            default_pool_size=pool_size,
            default_activation=activation,
            default_norm=norm,
            default_stride=stride,
            default_pool_stride=pool_stride,
            default_dropout=dropout,
            default_initializer=initializer,
            default_regularize=regularize
        )

        self.fc_stack = FCStack(
            layers=fc_layers,
            num_layers=num_fc_layers,
            default_fc_size=fc_size,
            default_activation=activation,
            default_norm=norm,
            default_dropout=dropout,
            default_regularize=regularize,
            default_initializer=initializer
        )

    def __call__(
            self,
            input_sequence,
            regularizer,
            dropout_rate,
            is_training=True
    ):
        """
            :param input_sequence: The input sequence fed into the encoder.
                   Shape: [batch x sequence length], type tf.int32
            :type input_sequence: Tensor
            :param regularizer: The regularizer to use for the weights
                   of the encoder.
            :type regularizer:
            :param dropout_rate: Tensor (tf.float) of the probability of dropout
            :type dropout_rate: Tensor
            :param is_training: Tesnor (tf.bool) specifying if in training mode
                   (important for dropout)
            :type is_training: Tensor
        """
        # ================ Embeddings ================
        if self.should_embed:
            embedded_input_sequence, self.embedding_size = self.embed_sequence(
                input_sequence,
                regularizer,
                dropout_rate,
                is_training=True
            )
        else:
            embedded_input_sequence = input_sequence
            while len(embedded_input_sequence.shape) < 3:
                embedded_input_sequence = tf.expand_dims(
                    embedded_input_sequence,
                    -1
                )
            self.embedding_size = embedded_input_sequence.shape[-1]

        hidden = embedded_input_sequence
        logging.debug('  hidden: {0}'.format(hidden))

        # ================ Conv Layers ================
        with tf.variable_scope('stack_parallel_conv'):
            hidden = self.stack_parallel_conv_1d(
                hidden,
                self.embedding_size,
                regularizer=regularizer,
                dropout_rate=dropout_rate,
                is_training=is_training
            )
        hidden_size = 0
        for stack in self.stacked_layers:
            hidden_size += stack[-1]['num_filters']
        logging.debug('  hidden: {0}'.format(hidden))

        # ================ Sequence Reduction ================
        if self.reduce_output is not None:
            hidden = reduce_sequence(hidden, self.reduce_output)

            # ================ FC Layers ================
            hidden_size = hidden.shape.as_list()[-1]
            logging.debug('  flatten hidden: {0}'.format(hidden))

            hidden = self.fc_stack(
                hidden,
                hidden_size,
                regularizer=regularizer,
                dropout_rate=dropout_rate,
                is_training=is_training
            )
            hidden_size = hidden.shape.as_list()[-1]

        return hidden, hidden_size


class RNN:

    def __init__(
            self,
            should_embed=True,
            vocab=None,
            representation='dense',
            embedding_size=256,
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=False,
            num_layers=1,
            state_size=256,
            cell_type='rnn',
            bidirectional=False,
            dropout=False,
            initializer=None,
            regularize=True,
            reduce_output='last',
            **kwargs
    ):
        """
            :param should_embed: If True the input sequence is expected
                   to be made of integers and will be mapped into embeddings
            :type should_embed: Boolean
            :param vocab: Vocabulary of the input feature to encode
            :type vocab: List
            :param representation: the possible values are `dense` and `sparse`.
                   `dense` means the mebeddings are initialized randomly,
                   `sparse` meanse they are initialized to be one-hot encodings.
            :type representation: Str (one of 'dense' or 'sparse')
            :param embedding_size: it is the maximum embedding size, the actual
                   size will be `min(vocaularyb_size, embedding_size)`
                   for `dense` representations and exacly `vocaularyb_size`
                   for the `sparse` encoding, where `vocabulary_size` is
                   the number of different strings appearing in the training set
                   in the column the feature is named after (plus 1 for `<UNK>`).
            :type embedding_size: Integer
            :param embeddings_trainable: If `True` embeddings are trained during
                   the training process, if `False` embeddings are fixed.
                   It may be useful when loading pretrained embeddings
                   for avoiding finetuning them. This parameter has effect only
                   for `representation` is `dense` as `sparse` one-hot encodings
                    are not trainable.
            :type embeddings_trainable: Boolean
            :param pretrained_embeddings: by default `dense` embeddings
                   are initialized randomly, but this parameter allow to specify
                   a path to a file containing embeddings in the GloVe format.
                   When the file containing the embeddings is loaded, only the
                   embeddings with labels present in the vocabulary are kept,
                   the others are discarded. If the vocabulary contains strings
                   that have no match in the embeddings file, their embeddings
                   are initialized with the average of all other embedding plus
                   some random noise to make them different from each other.
                   This parameter has effect only if `representation` is `dense`.
            :type pretrained_embeddings: str (filepath)
            :param embeddings_on_cpu: by default embedings matrices are stored
                   on GPU memory if a GPU is used, as it allows
                   for faster access, but in some cases the embedding matrix
                   may be really big and this parameter forces the placement
                   of the embedding matrix in regular memroy and the CPU is used
                   to resolve them, slightly slowing down the process
                   as a result of data transfer between CPU and GPU memory.
            :param conv_layers: it is a list of dictionaries containing
                   the parameters of all the convolutional layers. The length
                   of the list determines the number of parallel convolutional
                   layers and the content of each dictionary determines
                   the parameters for a specific layer. The available parameters
                   for each layer are: `filter_size`, `num_filters`, `pool`,
                   `norm`, `activation` and `regularize`. If any of those values
                   is missing from the dictionary, the default one specified
                   as a parameter of the encoder will be used instead. If both
                   `conv_layers` and `num_conv_layers` are `None`, a default
                   list will be assigned to `conv_layers` with the value
                   `[{filter_size: 2}, {filter_size: 3}, {filter_size: 4},
                   {filter_size: 5}]`.
            :type conv_layers: List
            :param num_conv_layers: if `conv_layers` is `None`, this is
                   the number of stacked convolutional layers.
            :type num_conv_layers: Integer
            :param filter_size:  if a `filter_size` is not already specified in
                   `conv_layers` this is the default `filter_size` that
                   will be used for each layer. It indicates how wide is
                   the 1d convolutional filter.
            :type filter_size: Integer
            :param num_filters: if a `num_filters` is not already specified in
                   `conv_layers` this is the default `num_filters` that
                   will be used for each layer. It indicates the number
                   of filters, and by consequence the output channels of
                   the 1d convolution.
            :type num_filters: Integer
            :param pool_size: if a `pool_size` is not already specified
                  in `conv_layers` this is the default `pool_size` that
                  will be used for each layer. It indicates the size of
                  the max pooling that will be performed along the `s` sequence
                  dimension after the convolution operation.
            :type pool_size: Integer
            :param num_rec_layers: the number of stacked recurrent layers.
            :type num_rec_layers: Integer
            :param cell_type: the type of recurrent cell to use.
                   Avalable values are: `rnn`, `lstm`, `lstm_block`, `lstm`,
                   `ln`, `lstm_cudnn`, `gru`, `gru_block`, `gru_cudnn`.
                   For reference about the differences between the cells please
                   refer to TensorFlow's documentstion. We suggest to use the
                   `block` variants on CPU and the `cudnn` variants on GPU
                   because of their increased speed.
            :type cell_type: str
            :param state_size: the size of the state of the rnn.
            :type state_size: Integer
            :param bidirectional: if `True` two recurrent networks will perform
                   encoding in the forward and backward direction and
                   their outputs will be concatenated.
            :type bidirectional: Boolean
            :param dropout: determines if there should be a dropout layer before
                   returning the encoder output.
            :type dropout: Boolean
            :param initializer: the initializer to use. If `None` it uses
                   `glorot_uniform`. Options are: `constant`, `identity`,
                   `zeros`, `ones`, `orthogonal`, `normal`, `uniform`,
                   `truncated_normal`, `variance_scaling`, `glorot_normal`,
                   `glorot_uniform`, `xavier_normal`, `xavier_uniform`,
                   `he_normal`, `he_uniform`, `lecun_normal`, `lecun_uniform`.
                   Alternatively it is possible to specify a dictionary with
                   a key `type` that identifies the type of initialzier and
                   other keys for its parameters,
                   e.g. `{type: normal, mean: 0, stddev: 0}`.
                   To know the parameters of each initializer, please refer
                   to TensorFlow's documentation.
            :type initializer: str
            :param regularize: if a `regularize` is not already specified in
                   `conv_layers` or `fc_layers` this is the default `regularize`
                   that will be used for each layer. It indicates if
                   the layer weights should be considered when computing
                   a regularization loss.
            :type regularize:
            :param reduce_output: defines how to reduce the output tensor of
                   the convolutional layers along the `s` sequence length
                   dimention if the rank of the tensor is greater than 2.
                   Available values are: `sum`, `mean` or `avg`, `max`, `concat`
                   (concatenates along the first dimension), `last` (returns
                   the last vector of the first dimension) and `None` or `null`
                   (which does not reduce and returns the full tensor).
            :type reduce_output: str
        """
        self.should_embed = should_embed

        self.embed_sequence = EmbedSequence(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            initializer=initializer,
            regularize=regularize
        )

        self.recurrent_stack = RecurrentStack(
            state_size=state_size,
            cell_type=cell_type,
            num_layers=num_layers,
            bidirectional=bidirectional,
            dropout=dropout,
            regularize=regularize,
            reduce_output=reduce_output
        )

    def __call__(
            self,
            input_sequence,
            regularizer,
            dropout_rate,
            is_training=True
    ):
        """
            :param input_sequence: The input sequence fed into the encoder.
                   Shape: [batch x sequence length], type tf.int32
            :type input_sequence: Tensor
            :param regularizer: The regularizer to use for the weights
                   of the encoder.
            :type regularizer:
            :param dropout_rate: Tensor (tf.float) of the probability of dropout
            :type dropout_rate: Tensor
            :param is_training: Tesnor (tf.bool) specifying if in training mode
                   (important for dropout)
            :type is_training: Tensor
        """
        # ================ Embeddings ================
        if self.should_embed:
            embedded_input_sequence, self.embedding_size = self.embed_sequence(
                input_sequence,
                regularizer,
                dropout_rate,
                is_training=True
            )
        else:
            embedded_input_sequence = input_sequence
            while len(embedded_input_sequence.shape) < 3:
                embedded_input_sequence = tf.expand_dims(
                    embedded_input_sequence,
                    -1
                )
            self.embedding_size = embedded_input_sequence.shape[-1]
        logging.debug('  hidden: {0}'.format(embedded_input_sequence))

        # ================ RNN ================
        hidden, hidden_size = self.recurrent_stack(
            embedded_input_sequence,
            regularizer=regularizer,
            dropout_rate=dropout_rate,
            is_training=is_training
        )

        return hidden, hidden_size


class CNNRNN:

    def __init__(
            self,
            should_embed=True,
            vocab=None,
            representation='dense',
            embedding_size=256,
            embeddings_trainable=True,
            pretrained_embeddings=None,
            embeddings_on_cpu=False,
            conv_layers=None,
            num_conv_layers=None,
            filter_size=5,
            num_filters=256,
            norm=None,
            activation='relu',
            pool_size=None,
            num_rec_layers=1,
            state_size=256,
            cell_type='rnn',
            bidirectional=False,
            dropout=False,
            initializer=None,
            regularize=True,
            reduce_output='last',
            **kwargs
    ):
        """
            :param should_embed: If True the input sequence is expected
                   to be made of integers and will be mapped into embeddings
            :type should_embed: Boolean
            :param vocab: Vocabulary of the input feature to encode
            :type vocab: List
            :param representation: the possible values are `dense` and `sparse`.
                   `dense` means the mebeddings are initialized randomly,
                   `sparse` meanse they are initialized to be one-hot encodings.
            :type representation: Str (one of 'dense' or 'sparse')
            :param embedding_size: it is the maximum embedding size, the actual
                   size will be `min(vocaularyb_size, embedding_size)`
                   for `dense` representations and exacly `vocaularyb_size`
                   for the `sparse` encoding, where `vocabulary_size` is
                   the number of different strings appearing in the training set
                   in the column the feature is named after (plus 1 for `<UNK>`).
            :type embedding_size: Integer
            :param embeddings_trainable: If `True` embeddings are trained during
                   the training process, if `False` embeddings are fixed.
                   It may be useful when loading pretrained embeddings
                   for avoiding finetuning them. This parameter has effect only
                   for `representation` is `dense` as `sparse` one-hot encodings
                    are not trainable.
            :type embeddings_trainable: Boolean
            :param pretrained_embeddings: by default `dense` embeddings
                   are initialized randomly, but this parameter allow to specify
                   a path to a file containing embeddings in the GloVe format.
                   When the file containing the embeddings is loaded, only the
                   embeddings with labels present in the vocabulary are kept,
                   the others are discarded. If the vocabulary contains strings
                   that have no match in the embeddings file, their embeddings
                   are initialized with the average of all other embedding plus
                   some random noise to make them different from each other.
                   This parameter has effect only if `representation` is `dense`.
            :type pretrained_embeddings: str (filepath)
            :param embeddings_on_cpu: by default embedings matrices are stored
                   on GPU memory if a GPU is used, as it allows
                   for faster access, but in some cases the embedding matrix
                   may be really big and this parameter forces the placement
                   of the embedding matrix in regular memroy and the CPU is used
                   to resolve them, slightly slowing down the process
                   as a result of data transfer between CPU and GPU memory.
            :param num_layers: the number of stacked recurrent layers.
            :type num_layers: Integer
            :param cell_type: the type of recurrent cell to use.
                   Avalable values are: `rnn`, `lstm`, `lstm_block`, `lstm`,
                   `ln`, `lstm_cudnn`, `gru`, `gru_block`, `gru_cudnn`.
                   For reference about the differences between the cells please
                   refer to TensorFlow's documentstion. We suggest to use the
                   `block` variants on CPU and the `cudnn` variants on GPU
                   because of their increased speed.
            :type cell_type: str
            :param state_size: the size of the state of the rnn.
            :type state_size: Integer
            :param bidirectional: if `True` two recurrent networks will perform
                   encoding in the forward and backward direction and
                   their outputs will be concatenated.
            :type bidirectional: Boolean
            :param dropout: determines if there should be a dropout layer before
                   returning the encoder output.
            :type dropout: Boolean
            :param initializer: the initializer to use. If `None` it uses
                   `glorot_uniform`. Options are: `constant`, `identity`,
                   `zeros`, `ones`, `orthogonal`, `normal`, `uniform`,
                   `truncated_normal`, `variance_scaling`, `glorot_normal`,
                   `glorot_uniform`, `xavier_normal`, `xavier_uniform`,
                   `he_normal`, `he_uniform`, `lecun_normal`, `lecun_uniform`.
                   Alternatively it is possible to specify a dictionary with
                   a key `type` that identifies the type of initialzier and
                   other keys for its parameters,
                   e.g. `{type: normal, mean: 0, stddev: 0}`.
                   To know the parameters of each initializer, please refer
                   to TensorFlow's documentation.
            :type initializer: str
            :param regularize: if a `regularize` is not already specified in
                   `conv_layers` or `fc_layers` this is the default `regularize`
                   that will be used for each layer. It indicates if
                   the layer weights should be considered when computing
                   a regularization loss.
            :type regularize:
            :param reduce_output: defines how to reduce the output tensor of
                   the convolutional layers along the `s` sequence length
                   dimention if the rank of the tensor is greater than 2.
                   Available values are: `sum`, `mean` or `avg`, `max`, `concat`
                   (concatenates along the first dimension), `last` (returns
                   the last vector of the first dimension) and `None` or `null`
                   (which does not reduce and returns the full tensor).
            :type reduce_output: str
        """
        if conv_layers is not None and num_conv_layers is None:
            # use custom-defined layers
            self.conv_layers = conv_layers
            self.num_conv_layers = len(conv_layers)
        elif conv_layers is None and num_conv_layers is not None:
            # generate num_conv_layers with default parameters
            self.conv_layers = None
            self.num_conv_layers = num_conv_layers
        elif conv_layers is None and num_conv_layers is None:
            # use default layers with varying filter sizes
            self.conv_layers = [
                {'pool_size': 3},
                {'pool_size': None}
            ]
            self.num_conv_layers = 2
        else:
            raise ValueError(
                'Invalid layer parametrization, use either conv_layers or num_conv_layers')

        self.should_embed = should_embed

        self.embed_sequence = EmbedSequence(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            initializer=initializer,
            regularize=regularize
        )

        self.conv_stack_1d = ConvStack1D(
            layers=self.conv_layers,
            default_filter_size=filter_size,
            default_num_filters=num_filters,
            default_activation=activation,
            default_norm=norm,
            default_pool_size=pool_size,
            default_dropout=dropout,
            default_initializer=initializer,
            default_regularize=regularize
        )

        self.recurrent_stack = RecurrentStack(
            state_size=state_size,
            cell_type=cell_type,
            num_layers=num_rec_layers,
            bidirectional=bidirectional,
            dropout=dropout,
            regularize=regularize,
            reduce_output=reduce_output
        )

    def __call__(
            self,
            input_sequence,
            regularizer,
            dropout_rate,
            is_training=True
    ):
        """
            :param input_sequence: The input sequence fed into the encoder.
                   Shape: [batch x sequence length], type tf.int32
            :type input_sequence: Tensor
            :param regularizer: The regularizer to use for the weights
                   of the encoder.
            :type regularizer:
            :param dropout_rate: Tensor (tf.float) of the probability of dropout
            :type dropout_rate: Tensor
            :param is_training: Tesnor (tf.bool) specifying if in training mode
                   (important for dropout)
            :type is_training: Tensor
        """
        # ================ Embeddings ================
        if self.should_embed:
            embedded_input_sequence, self.embedding_size = self.embed_sequence(
                input_sequence,
                regularizer,
                dropout_rate,
                is_training=True
            )
        else:
            embedded_input_sequence = input_sequence
            while len(embedded_input_sequence.shape) < 3:
                embedded_input_sequence = tf.expand_dims(
                    embedded_input_sequence,
                    -1
                )
            self.embedding_size = embedded_input_sequence.shape[-1]

        hidden = embedded_input_sequence
        # shape=(?, sequence_length, embedding_size)
        logging.debug('  hidden: {0}'.format(hidden))

        # ================ CNN ================
        hidden = self.conv_stack_1d(
            hidden,
            self.embedding_size,
            regularizer=regularizer,
            dropout_rate=dropout_rate,
            is_training=is_training
        )
        logging.debug('  hidden: {0}'.format(hidden))

        # ================ RNN ================
        hidden, hidden_size = self.recurrent_stack(
            hidden,
            regularizer=regularizer,
            dropout_rate=dropout_rate,
            is_training=is_training
        )

        return hidden, hidden_size



# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tensorflow as tf

from ludwig.utils.misc import get_from_registry

initializers_registry = {
    'constant': tf.initializers.constant,
    'identity': tf.initializers.identity,
    'zeros': tf.initializers.zeros,
    'ones': tf.initializers.ones,
    'orthogonal': tf.initializers.orthogonal,
    'normal': tf.initializers.random_normal,
    'uniform': tf.initializers.random_uniform,
    'truncated_normal': tf.initializers.truncated_normal,
    'variance_scaling': tf.initializers.variance_scaling,
    'glorot_normal': tf.initializers.glorot_normal,
    'glorot_uniform': tf.initializers.glorot_uniform,
    'xavier_normal': tf.initializers.glorot_normal,
    'xavier_uniform': tf.initializers.glorot_uniform,
    'he_normal': tf.initializers.he_normal,
    'he_uniform': tf.initializers.he_uniform,
    'lecun_normal': tf.initializers.lecun_normal,
    'lecun_uniform': tf.initializers.lecun_uniform,
    None: tf.initializers.glorot_uniform
}


def get_initializer(parameters):
    if parameters is None:
        return initializers_registry[parameters]()
    elif isinstance(parameters, str):
        initializer_fun = get_from_registry(
            parameters, initializers_registry)
        return initializer_fun()
    elif isinstance(parameters, dict):
        initializer_fun = get_from_registry(
            parameters['type'], initializers_registry)
        arguments = parameters.copy()
        del arguments['type']
        return initializer_fun(**arguments)
    else:
        raise ValueError(
            'Initializers parameters should be either strings or dictionaries, '
            'but the provided parameters are a {}. '
            'Parameters values: {}'.format(
                type(parameters), parameters
            ))



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""
This module contains the class and auxiliary methods of a model.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import json
import logging
import os
import re
import signal
import sys
import time
from collections import OrderedDict

import numpy as np
import tensorflow as tf
from tabulate import tabulate
from tensorflow.python import debug as tf_debug
from tqdm import tqdm

from ludwig.constants import *
from ludwig.features.feature_registries import output_type_registry
from ludwig.features.feature_utils import SEQUENCE_TYPES
from ludwig.globals import MODEL_WEIGHTS_FILE_NAME
from ludwig.globals import MODEL_HYPERPARAMETERS_FILE_NAME
from ludwig.globals import TRAINING_PROGRESS_FILE_NAME
from ludwig.globals import MODEL_WEIGHTS_PROGRESS_FILE_NAME
from ludwig.globals import is_progressbar_disabled
from ludwig.globals import is_on_master
from ludwig.models.combiners import get_build_combiner
from ludwig.models.inputs import build_inputs, dynamic_length_encoders
from ludwig.models.modules.loss_modules import regularizer_registry
from ludwig.models.modules.measure_modules import get_initial_validation_value
from ludwig.models.modules.measure_modules import get_improved_fun
from ludwig.models.modules.optimization_modules import optimize
from ludwig.models.outputs import build_outputs
from ludwig.utils import time_utils
from ludwig.utils.batcher import Batcher
from ludwig.utils.batcher import BucketedBatcher
from ludwig.utils.batcher import DistributedBatcher
from ludwig.utils.data_utils import load_json
from ludwig.utils.data_utils import load_object
from ludwig.utils.data_utils import save_object
from ludwig.utils.defaults import default_random_seed
from ludwig.utils.defaults import default_training_params
from ludwig.utils.math_utils import learning_rate_warmup
from ludwig.utils.misc import set_random_seed
from ludwig.utils.misc import sum_dicts
from ludwig.utils.tf_utils import get_tf_config


class Model:
    """
    Model is a class that builds the model that Ludwig uses
    """

    def __init__(
            self,
            input_features,
            output_features,
            combiner,
            training,
            preprocessing,
            use_horovod=False,
            random_seed=default_random_seed,
            debug=False,
            **kwargs
    ):
        self.horovod = None
        if use_horovod:
            import horovod.tensorflow
            self.horovod = horovod.tensorflow
            from mpi4py import MPI
            self.comm = MPI.COMM_WORLD

        self.debug = debug
        self.weights_save_path = None
        self.hyperparameters = {}
        self.session = None

        self.epochs = None
        self.received_sigint = False

        self.__build(
            input_features,
            output_features,
            combiner,
            training,
            preprocessing,
            random_seed,
            **kwargs
        )

    def __build(
            self,
            input_features,
            output_features,
            combiner,
            training,
            preprocessing,
            random_seed,
            **kwargs
    ):
        self.hyperparameters['input_features'] = input_features
        self.hyperparameters['output_features'] = output_features
        self.hyperparameters['combiner'] = combiner
        self.hyperparameters['training'] = training
        self.hyperparameters['preprocessing'] = preprocessing
        self.hyperparameters['random_seed'] = random_seed
        self.hyperparameters.update(kwargs)

        if self.horovod:
            self.horovod.init()

        tf.reset_default_graph()
        graph = tf.Graph()
        with graph.as_default():
            # ================ Setup ================
            tf.set_random_seed(random_seed)

            self.global_step = tf.Variable(0, trainable=False)
            self.regularization_lambda = tf.placeholder(
                tf.float32,
                name='regularization_lambda'
            )
            regularizer = regularizer_registry[training['regularizer']]
            self.regularizer = regularizer(self.regularization_lambda)

            self.learning_rate = tf.placeholder(
                tf.float32,
                name='learning_rate'
            )
            self.dropout_rate = tf.placeholder(tf.float32, name='dropout_rate')
            self.is_training = tf.placeholder(tf.bool, [], name='is_training')

            # ================ Inputs ================
            feature_encodings = build_inputs(
                input_features,
                self.regularizer,
                self.dropout_rate,
                is_training=self.is_training
            )

            for fe_name, fe_properties in feature_encodings.items():
                setattr(self, fe_name, fe_properties['placeholder'])

            # ================ Model ================
            logging.debug('- Combiner {}'.format(combiner['type']))
            build_combiner = get_build_combiner(combiner['type'])(**combiner)
            hidden, hidden_size = build_combiner(
                feature_encodings,
                self.regularizer,
                self.dropout_rate,
                is_training=self.is_training,
                **kwargs
            )

            # ================ Outputs ================
            outs = build_outputs(
                output_features,
                hidden,
                hidden_size,
                regularizer=self.regularizer,
                dropout_rate=self.dropout_rate,
                is_training=self.is_training
            )

            (
                self.train_reg_mean_loss,
                self.eval_combined_loss,
                self.regularization_loss,
                output_tensors
            ) = outs

            for ot_name, ot in output_tensors.items():
                setattr(self, ot_name, ot)

            # ================ Optimizer ================
            self.optimize, self.learning_rate = optimize(
                self.train_reg_mean_loss,
                training,
                self.learning_rate,
                self.global_step,
                self.horovod
            )

            tf.summary.scalar('train_reg_mean_loss', self.train_reg_mean_loss)

            self.merged = tf.summary.merge_all()
            self.graph = graph
            self.graph_initialize = tf.global_variables_initializer()
            if self.horovod:
                self.broadcast_op = self.horovod.broadcast_global_variables(0)
            self.saver = tf.train.Saver()

    def initialize_session(self, gpus=None, gpu_fraction=1):
        if self.session is None:

            self.session = tf.Session(
                config=get_tf_config(gpus, gpu_fraction, self.horovod),
                graph=self.graph
            )
            self.session.run(self.graph_initialize)

            if self.debug:
                session = tf_debug.LocalCLIDebugWrapperSession(self.session)
                session.add_tensor_filter(
                    'has_inf_or_nan',
                    tf_debug.has_inf_or_nan
                )

        return self.session

    def close_session(self):
        if self.session is not None:
            self.session.close()
            self.session = None

    def feed_dict(
            self,
            batch,
            regularization_lambda=default_training_params['regularization_lambda'],
            learning_rate=default_training_params['learning_rate'],
            dropout_rate=default_training_params['dropout_rate'],
            is_training=True
    ):
        input_features = self.hyperparameters['input_features']
        output_features = self.hyperparameters['output_features']
        feed_dict = {
            self.is_training: is_training,
            self.regularization_lambda: regularization_lambda,
            self.learning_rate: learning_rate,
            self.dropout_rate: dropout_rate
        }
        for input_feature in input_features:
            feed_dict[getattr(self, input_feature['name'])] = batch[
                input_feature['name']]
        for output_feature in output_features:
            if output_feature['name'] in batch:
                feed_dict[getattr(self, output_feature['name'])] = batch[
                    output_feature['name']]
        return feed_dict

    def train(
            self,
            training_set,
            validation_set=None,
            test_set=None,
            validation_field=None,
            validation_measure=None,
            save_path='model',
            regularization_lambda=0.0,
            epochs=100,
            learning_rate=0.001,
            batch_size=128,
            bucketing_field=None,
            dropout_rate=0.0,
            early_stop=20,
            reduce_learning_rate_on_plateau=0,
            reduce_learning_rate_on_plateau_patience=5,
            reduce_learning_rate_on_plateau_rate=0.5,
            increase_batch_size_on_plateau=0,
            increase_batch_size_on_plateau_patience=5,
            increase_batch_size_on_plateau_rate=2,
            increase_batch_size_on_plateau_max=512,
            learning_rate_warmup_epochs=5,  # used when training with Horovod
            resume=False,
            skip_save_progress_weights=False,
            gpus=None,
            gpu_fraction=1,
            random_seed=default_random_seed,
            **kwargs
    ):
        """Trains a model with a set of hyperparameters listed below. Customizable
        :param training_set: The training set
        :param validation_set: The validation dataset
        :param test_set: The test dataset
        :param validation_field: The first output feature, by default it is set
               as the same field of the first output feature.
        :param validation_measure: Measure used on the validation field, it is
               accuracy by default
        :type validation_measure:
        :param save_path: The path to save the file
        :type save_path: filepath (str)
        :param regularization_lambda: Strength of the $L2$ regularization
        :type regularization_lambda: Integer
        :param epochs: Number of epochs the algorithm is intended to be run over
        :type epochs: Integer
        :param learning_rate: Learning rate for the algorithm, represents how
               much to scale the gradients by
        :type learning_rate: Integer
        :param batch_size: Size of batch to pass to the machine learning model.
        :type batch_size: Integer
        :param bucketing_field:when batching, buckets datapoints based the
               length of a field together. Bucketing on text length speeds up
               training of RNNs consistently, 30% in some cases
        :type bucketing_field:
        :param dropout_rate: dropout_rate probability (probability of dropping
               a neuron in a given layer)
        :type dropout_rate: Float
        :param early_stop: How many epochs without any improvement in the
               validation_measure triggers the algorithm to stop
        :type early_stop: Integer
        :param reduce_learning_rate_on_plateau: Reduces the learning rate when
               the algorithm hits a plateau (i.e. the performance on the
               validation doesn't improve)
        :type reduce_learning_rate_on_plateau: Float
        :param reduce_learning_rate_on_plateau_patience: How many epochs have
               to pass before the learning rate reduces
        :type reduce_learning_rate_on_plateau_patience: Float
        :param reduce_learning_rate_on_plateau_rate: Rate at which we reduce
               the learning rate
        :type reduce_learning_rate_on_plateau_rate: Float
        :param increase_batch_size_on_plateau: Increase the batch size on a
               plateau
        :type increase_batch_size_on_plateau: Integer
        :param increase_batch_size_on_plateau_patience: How many epochs to wait
               for before increasing the batch size
        :type increase_batch_size_on_plateau_patience: Integer
        :param increase_batch_size_on_plateau_rate: The rate at which the batch
               size increases.
        :type increase_batch_size_on_plateau_rate: Float
        :param increase_batch_size_on_plateau_max: The maximum size of the batch
        :type increase_batch_size_on_plateau_max: Integer
        :param learning_rate_warmup_epochs: The number of epochs to warmup the
               learning rate for.
        :type learning_rate_warmup_epochs: Integer
        :param resume: Resume training a model that was being trained.
        :type resume: Boolean
        :param skip_save_progress_weights:
        :type skip_save_progress_weights:
        :param gpus: List of gpus to use
        :type gpus: List
        :param gpu_fraction: Percentage of the GPU that is intended to be used
        :type gpu_fraction: Float
        :param random_seed: Default initialization for the random seeds
        :type: Float
        """
        # ====== General setup =======
        output_features = self.hyperparameters['output_features']
        self.epochs = epochs
        digits_per_epochs = len(str(self.epochs))
        self.received_sigint = False
        signal.signal(signal.SIGINT, self.set_epochs_to_1_or_quit)
        should_validate = validation_set is not None and validation_set.size > 0
        stat_names = self.get_stat_names(output_features)
        if self.horovod:
            learning_rate *= self.horovod.size()

        # ====== Setup file names =======
        model_weights_path = os.path.join(save_path, MODEL_WEIGHTS_FILE_NAME)
        model_weights_progress_path = os.path.join(
            save_path,
            MODEL_WEIGHTS_PROGRESS_FILE_NAME
        )
        model_hyperparameters_path = os.path.join(
            save_path,
            MODEL_HYPERPARAMETERS_FILE_NAME
        )

        # ====== Setup session =======
        session = self.initialize_session(gpus, gpu_fraction)

        if is_on_master():
            train_writer = tf.summary.FileWriter(
                os.path.join(save_path, 'log', 'train'),
                session.graph
            )

        if self.debug:
            session = tf_debug.LocalCLIDebugWrapperSession(session)
            session.add_tensor_filter(
                'has_inf_or_nan',
                tf_debug.has_inf_or_nan
            )

        # ================ Resume logic ================
        if resume:
            progress_tracker = self.resume_training(
                save_path,
                model_weights_path
            )
            if is_on_master():
                self.resume_session(
                    session,
                    save_path,
                    model_weights_path,
                    model_weights_progress_path
                )
        else:
            (
                train_stats,
                vali_stats,
                test_stats
            ) = self.initialize_training_stats(output_features)

            progress_tracker = ProgressTracker(
                batch_size=batch_size,
                epoch=0,
                steps=0,
                last_improvement_epoch=0,
                learning_rate=learning_rate,
                best_valid_measure=get_initial_validation_value(
                    validation_measure
                ),
                num_reductions_lr=0,
                num_increases_bs=0,
                train_stats=train_stats,
                vali_stats=vali_stats,
                test_stats=test_stats
            )

        # horovod broadcasting after init or restore
        if self.horovod:
            session.run(self.broadcast_op)

        set_random_seed(random_seed)
        batcher = self.initialize_batcher(
            training_set,
            batch_size,
            bucketing_field
        )

        # ================ Training Loop ================
        while progress_tracker.epoch < self.epochs:
            # epoch init
            start_time = time.time()
            if is_on_master():
                logging.info(
                    '\nEpoch {epoch:{digits}d}'.format(
                        epoch=progress_tracker.epoch + 1,
                        digits=digits_per_epochs
                    )
                )
            # needed because batch size may change
            batcher.batch_size = progress_tracker.batch_size

            # ================ Train ================
            if is_on_master():
                bar = tqdm(
                    desc='Training',
                    total=batcher.steps_per_epoch,
                    file=sys.stdout,
                    disable=is_progressbar_disabled()
                )

            # training step loop
            while not batcher.last_batch():
                batch = batcher.next_batch()

                if self.horovod:
                    current_learning_rate = learning_rate_warmup(
                        progress_tracker.learning_rate,
                        progress_tracker.epoch,
                        learning_rate_warmup_epochs,
                        self.horovod.size(),
                        batcher.steps_per_epoch
                    )
                else:
                    current_learning_rate = progress_tracker.learning_rate

                summary, _ = session.run(
                    [self.merged, self.optimize],
                    feed_dict=self.feed_dict(
                        batch,
                        regularization_lambda=regularization_lambda,
                        learning_rate=current_learning_rate,
                        dropout_rate=dropout_rate,
                        is_training=True
                    )
                )
                if is_on_master():
                    # it is initialized only on master
                    train_writer.add_summary(summary, progress_tracker.steps)

                progress_tracker.steps += 1
                if is_on_master():
                    bar.update(1)

            # post training
            if is_on_master():
                bar.close()

            progress_tracker.epoch += 1
            batcher.reset()  # todo this may be useless, doublecheck

            # ================ Eval ================
            # init tables
            tables = OrderedDict()
            for output_feature in output_features:
                field_name = output_feature['name']
                tables[field_name] = [
                    [field_name] + stat_names[field_name]]
            tables['combined'] = [['combined', LOSS, ACCURACY]]

            # eval measures on train set
            self.evaluation(
                session,
                training_set,
                'train',
                regularization_lambda,
                progress_tracker.train_stats,
                tables,
                progress_tracker.batch_size,
                bucketing_field
            )

            # eval measures on validation set
            self.evaluation(
                session,
                validation_set,
                'vali',
                regularization_lambda,
                progress_tracker.vali_stats,
                tables,
                progress_tracker.batch_size,
                bucketing_field
            )

            # eval measures on test set
            self.evaluation(
                session,
                test_set,
                'test',
                regularization_lambda,
                progress_tracker.test_stats,
                tables,
                progress_tracker.batch_size,
                bucketing_field
            )

            # mbiu and end of epoch prints
            elapsed_time = (time.time() - start_time) * 1000.0

            if is_on_master():
                logging.info('Took {time}'.format(
                    time=time_utils.strdelta(elapsed_time)))

            # stat prints
            for output_feature, table in tables.items():
                if (
                        output_feature != 'combined' or
                        (output_feature == 'combined' and
                         len(output_features) > 1)
                ):
                    if is_on_master():
                        logging.info(
                            tabulate(
                                table,
                                headers='firstrow',
                                tablefmt='fancy_grid',
                                floatfmt='.4f'
                            )
                        )

            if should_validate:
                should_break = self.check_progress_on_validation(
                    progress_tracker,
                    validation_field,
                    validation_measure,
                    session,
                    model_weights_path,
                    model_hyperparameters_path,
                    reduce_learning_rate_on_plateau,
                    reduce_learning_rate_on_plateau_patience,
                    reduce_learning_rate_on_plateau_rate,
                    increase_batch_size_on_plateau_patience,
                    increase_batch_size_on_plateau,
                    increase_batch_size_on_plateau_max,
                    increase_batch_size_on_plateau_rate,
                    early_stop
                )
                if should_break:
                    break
            else:
                # there's no validation, so we save the model at each iteration
                if is_on_master():
                    self.save_weights(session, model_weights_path)
                    self.save_hyperparameters(
                        self.hyperparameters,
                        model_hyperparameters_path
                    )

            # ========== Save training progress ==========
            if is_on_master():
                save_object(
                    os.path.join(
                        save_path,
                        TRAINING_PROGRESS_FILE_NAME
                    ),
                    progress_tracker
                )
                if not skip_save_progress_weights:
                    self.save_weights(session, model_weights_progress_path)

            if is_on_master():
                logging.info('')

        return (
            progress_tracker.train_stats,
            progress_tracker.vali_stats,
            progress_tracker.test_stats
        )

    def train_online(
            self,
            dataset,
            batch_size=128,
            learning_rate=0.01,
            regularization_lambda=0,
            dropout_rate=0,
            bucketing_field=None,
            gpus=None,
            gpu_fraction=1
    ):
        session = self.initialize_session(gpus, gpu_fraction)
        batcher = self.initialize_batcher(dataset, batch_size, bucketing_field)

        # training step loop
        bar = tqdm(
            desc='Trainining online',
            total=batcher.steps_per_epoch,
            file=sys.stdout,
            disable=is_progressbar_disabled()
        )

        while not batcher.last_batch():
            batch = batcher.next_batch()

            _ = session.run(
                [self.optimize],
                feed_dict=self.feed_dict(
                    batch,
                    regularization_lambda=regularization_lambda,
                    learning_rate=learning_rate,
                    dropout_rate=dropout_rate,
                    is_training=True
                )
            )
            bar.update(1)

        bar.close()

    def evaluation(
            self,
            session,
            dataset,
            dataset_name,
            regularization_lambda,
            stats,
            tables,
            batch_size=128,
            bucketing_field=None
    ):
        results = self.batch_evaluation(
            session,
            dataset,
            batch_size,
            bucketing_field=bucketing_field,
            regularization_lambda=regularization_lambda,
            is_training=False,
            name=dataset_name
        )

        for output_feature in self.hyperparameters['output_features']:
            field_name = output_feature['name']
            scores = [dataset_name]

            for stat in stats[field_name]:
                stats[field_name][stat].append(results[field_name][stat])
                scores.append(results[field_name][stat])

            tables[field_name].append(scores)

        stats['combined'][LOSS].append(results['combined'][LOSS])
        stats['combined'][ACCURACY].append(results['combined'][ACCURACY])
        tables['combined'].append(
            [
                dataset_name,
                results['combined'][LOSS],
                results['combined'][ACCURACY]
            ]
        )
        return stats, tables

    def batch_evaluation(
            self,
            session,
            dataset,
            batch_size,
            bucketing_field=None,
            regularization_lambda=0.0,
            is_training=False,
            collect_predictions=False,
            only_predictions=False,
            name=None
    ):
        output_nodes = self.get_output_nodes(
            collect_predictions,
            only_predictions
        )
        output_stats = self.get_outputs_stats()

        set_size = dataset.size
        if set_size == 0:
            if is_on_master():
                logging.warning('No datapoints to evaluate on.')
            return output_stats
        seq_set_size = {output_feature['name']: {} for output_feature in
                        self.hyperparameters['output_features'] if
                        output_feature['type'] in SEQUENCE_TYPES}

        batcher = self.initialize_batcher(
            dataset,
            batch_size,
            bucketing_field,
            should_shuffle=False
        )

        if is_on_master():
            bar = tqdm(
                desc='Evaluation' if name is None
                else 'Evaluation {0: <5.5}'.format(name),
                total=batcher.steps_per_epoch,
                file=sys.stdout,
                disable=is_progressbar_disabled()
            )

        while not batcher.last_batch():
            batch = batcher.next_batch()
            result = session.run(
                output_nodes,
                feed_dict=self.feed_dict(
                    batch,
                    regularization_lambda=regularization_lambda,
                    dropout_rate=0.0,
                    is_training=is_training
                )
            )

            output_stats, seq_set_size = self.update_output_stats_batch(
                output_stats,
                seq_set_size,
                collect_predictions,
                only_predictions,
                result
            )
            if is_on_master():
                bar.update(1)

        if is_on_master():
            bar.close()

        if self.horovod:
            output_stats, seq_set_size = self.merge_workers_outputs(
                output_stats,
                seq_set_size
            )

        output_stats = self.update_output_stats(
            output_stats,
            set_size,
            seq_set_size,
            collect_predictions,
            only_predictions
        )

        if 'combined' in output_stats and LOSS in output_stats['combined']:
            regularization = session.run(
                [self.regularization_loss],
                feed_dict={self.regularization_lambda: regularization_lambda}
            )[0]
            output_stats['combined'][LOSS] += regularization

        return output_stats

    def merge_workers_outputs(self, output_stats, seq_set_size):
        # gather outputs from all workers
        all_workers_output_stats = self.comm.allgather(output_stats)
        all_workers_seq_set_size = self.comm.allgather(seq_set_size)

        # merge them into a single one
        merged_output_stats = sum_dicts(
            all_workers_output_stats,
            dict_type=OrderedDict
        )
        merged_seq_set_size = sum_dicts(all_workers_seq_set_size)

        return merged_output_stats, merged_seq_set_size

    def batch_collect_activations(
            self,
            session,
            dataset,
            batch_size,
            tensor_names,
            bucketing_field=None
    ):
        output_nodes = {tensor_name: self.graph.get_tensor_by_name(tensor_name)
                        for tensor_name in tensor_names}
        collected_tensors = {tensor_name: [] for tensor_name in tensor_names}

        batcher = self.initialize_batcher(
            dataset,
            batch_size,
            bucketing_field,
            should_shuffle=False
        )

        bar = tqdm(
            desc='Collecting Tensors',
            total=batcher.steps_per_epoch,
            file=sys.stdout,
            disable=is_progressbar_disabled()
        )

        while not batcher.last_batch():
            batch = batcher.next_batch()
            result = session.run(
                output_nodes,
                feed_dict=self.feed_dict(
                    batch,
                    is_training=False
                )
            )

            for tensor_name in result:
                for row in result[tensor_name]:
                    collected_tensors[tensor_name].append(row)

            bar.update(1)

        bar.close()

        return collected_tensors

    def get_output_nodes(self, collect_predictions, only_predictions=False):
        output_features = self.hyperparameters['output_features']
        output_nodes = {}

        for output_feature in output_features:
            field_name = output_feature['name']
            feature_type = output_feature['type']
            output_nodes[field_name] = {}
            output_config = output_type_registry[feature_type].output_config

            for stat in output_config:
                output_name = output_config[stat]['output']
                output_type = output_config[stat]['type']
                if ((output_type == PREDICTION and
                     (collect_predictions or only_predictions)) or
                        (output_type == MEASURE and not only_predictions)):
                    output_nodes[field_name][output_name] = getattr(
                        self,
                        output_name + '_' + field_name
                    )

        if not only_predictions:
            output_nodes['eval_combined_loss'] = getattr(
                self,
                'eval_combined_loss'
            )

        return output_nodes

    def get_outputs_stats(self):
        output_features = self.hyperparameters['output_features']
        output_stats = OrderedDict()

        for output_feature in output_features:
            field_name = output_feature['name']
            feature_type = output_feature['type']
            output_stats[field_name] = {}
            output_config = output_type_registry[feature_type].output_config

            for stat in output_config:
                output_value = output_config[stat]['value']
                if isinstance(output_value, list):
                    output_stats[field_name][stat] = []
                else:
                    output_stats[field_name][stat] = output_value

        output_stats['combined'] = {LOSS: 0, ACCURACY: 0}
        return output_stats

    def update_output_stats_batch(
            self,
            output_stats,
            seq_set_size,
            collect_predictions,
            only_predictions,
            result
    ):
        output_features = self.hyperparameters['output_features']
        combined_correct_predictions = None

        for i, output_feature in enumerate(output_features):
            field_name = output_feature['name']
            feature_type = output_feature['type']
            output_config = output_type_registry[feature_type].output_config

            for stat in output_config:
                stat_config = output_config[stat]
                output_type = output_config[stat]['type']
                if ((output_type == PREDICTION and
                     (collect_predictions or only_predictions)) or
                        (output_type == MEASURE and not only_predictions)):
                    aggregation_method = stat_config['aggregation']
                    if aggregation_method == SUM:
                        output_stats[field_name][stat] += (
                            result[field_name][stat_config['output']].sum()
                        )
                    elif aggregation_method == SEQ_SUM:
                        output_stats[field_name][stat] += (
                            result[field_name][stat_config['output']].sum()
                        )
                        seq_set_size[field_name][stat] = (
                            seq_set_size[field_name].get(stat, 0) +
                            len(result[field_name][stat_config['output']])
                        )
                    elif aggregation_method == AVG_EXP:
                        output_stats[field_name][stat] += (
                            result[field_name][stat_config['output']].sum()
                        )
                    elif aggregation_method == APPEND:
                        output_stats[field_name][stat].append(
                            result[field_name][stat_config['output']]
                        )

            if not only_predictions:
                if feature_type in [CATEGORY, BINARY]:
                    correct_predictions = \
                        result[field_name][CORRECT_PREDICTIONS]
                elif feature_type == SEQUENCE:
                    correct_predictions = \
                        result[field_name][CORRECT_ROWWISE_PREDICTIONS]
                else:
                    correct_predictions = None

                if correct_predictions is not None:
                    if combined_correct_predictions is None:
                        combined_correct_predictions = correct_predictions
                    else:
                        combined_correct_predictions = np.logical_and(
                            combined_correct_predictions,
                            correct_predictions
                        )

        if not only_predictions:
            output_stats['combined'][LOSS] += result['eval_combined_loss'].sum()
            output_stats['combined'][ACCURACY] += (
                combined_correct_predictions.sum()
                if combined_correct_predictions is not None else 0
            )

        return output_stats, seq_set_size

    def update_output_stats(
            self,
            output_stats,
            set_size,
            seq_set_size,
            collect_predictions,
            only_predictions
    ):
        output_features = self.hyperparameters['output_features']

        for i, output_feature in enumerate(output_features):
            feature_type = output_feature['type']
            field_name = output_feature['name']
            output_config = output_type_registry[feature_type].output_config

            for stat in output_config:
                output_type = output_config[stat]['type']
                if ((output_type == PREDICTION and
                     (collect_predictions or only_predictions)) or
                        (output_type == MEASURE and not only_predictions)):

                    if output_config[stat]['aggregation'] == SUM:
                        output_stats[field_name][stat] /= set_size

                    elif output_config[stat]['aggregation'] == SEQ_SUM:
                        output_stats[field_name][stat] /= (
                            seq_set_size[field_name][stat]
                        )

                    elif output_config[stat]['aggregation'] == AVG_EXP:
                        output_stats[field_name][stat] = np.exp(
                            output_stats[field_name][stat] / set_size
                        )

                    elif output_config[stat]['aggregation'] == APPEND:
                        if len(output_stats[field_name][stat]) > 0 and len(
                                output_stats[field_name][stat][0].shape) > 1:
                            max_shape = None
                            for result in output_stats[field_name][stat]:
                                if max_shape is None:
                                    max_shape = result.shape
                                else:
                                    max_shape = np.maximum(
                                        max_shape,
                                        result.shape
                                    )

                            results = []
                            for result in output_stats[field_name][stat]:
                                diff_shape = max_shape - np.array(result.shape)
                                diff_shape[0] = 0
                                pad_width = [(0, k) for k in diff_shape]
                                paded_result = np.pad(
                                    result,
                                    pad_width,
                                    'constant',
                                    constant_values=0
                                )
                                results.append(paded_result)
                        else:
                            results = output_stats[field_name][stat]

                        output_stats[field_name][stat] = np.concatenate(
                            results
                        )

            if feature_type == SEQUENCE:
                # trim output sequences
                if LENGTHS in output_stats[field_name]:
                    lengths = output_stats[field_name][LENGTHS]
                    if PREDICTIONS in output_stats[field_name]:
                        output_stats[field_name][PREDICTIONS] = np.array(
                            [list(output_stats[field_name][PREDICTIONS][i,
                                  0:lengths[i]])
                             for i in range(len(lengths))]
                        )
                    if PROBABILITIES in output_stats[field_name]:
                        output_stats[field_name][PROBABILITIES] = np.array(
                            [list(output_stats[field_name][PROBABILITIES][i,
                                  0:lengths[i]]) for i in
                             range(len(lengths))]
                        )

        if not only_predictions:
            output_stats['combined'][LOSS] /= set_size
            output_stats['combined'][ACCURACY] /= set_size

        return output_stats

    def check_progress_on_validation(
            self,
            progress_tracker,
            validation_field,
            validation_measure,
            session, model_weights_path,
            model_hyperparameters_path,
            reduce_learning_rate_on_plateau,
            reduce_learning_rate_on_plateau_patience,
            reduce_learning_rate_on_plateau_rate,
            increase_batch_size_on_plateau_patience,
            increase_batch_size_on_plateau,
            increase_batch_size_on_plateau_max,
            increase_batch_size_on_plateau_rate,
            early_stop
    ):
        should_break = False
        # record how long its been since an improvement
        improved = get_improved_fun(validation_measure)
        if improved(
                progress_tracker.vali_stats[validation_field][
                    validation_measure][-1],
                progress_tracker.best_valid_measure
        ):
            progress_tracker.last_improvement_epoch = progress_tracker.epoch
            progress_tracker.best_valid_measure = progress_tracker.vali_stats[
                validation_field][validation_measure][-1]
            if is_on_master():
                self.save_weights(session, model_weights_path)
                self.save_hyperparameters(
                    self.hyperparameters,
                    model_hyperparameters_path
                )

        progress_tracker.last_improvement = (
                progress_tracker.epoch - progress_tracker.last_improvement_epoch
        )
        if progress_tracker.last_improvement == 0:
            if is_on_master():
                logging.info(
                    'Validation {} on {} improved, model saved'.format(
                        validation_measure,
                        validation_field
                    )
                )
        else:
            if is_on_master():
                logging.info(
                    'Last improvement of {} on {} happened '
                    '{} epoch{} ago'.format(
                        validation_measure,
                        validation_field,
                        progress_tracker.last_improvement,
                        '' if progress_tracker.last_improvement == 1 else 's'
                    )
                )

        # ========== Reduce Learning Rate Plateau logic ========
        if reduce_learning_rate_on_plateau > 0:
            self.reduce_learning_rate(
                progress_tracker,
                reduce_learning_rate_on_plateau,
                reduce_learning_rate_on_plateau_patience,
                reduce_learning_rate_on_plateau_rate
            )

        # ========== Increase Batch Size Plateau logic =========
        if increase_batch_size_on_plateau > 0:
            self.increase_batch_size(
                progress_tracker,
                increase_batch_size_on_plateau_patience,
                increase_batch_size_on_plateau,
                increase_batch_size_on_plateau_max,
                increase_batch_size_on_plateau_rate
            )

        # ========== Early Stop logic ==========
        if early_stop > 0:
            if progress_tracker.last_improvement >= early_stop:
                if is_on_master():
                    logging.info(
                        "\nEARLY STOPPING due to lack of validation improvement"
                        ", it has been {0} epochs since last validation "
                        "accuracy improvement\n".format(
                            progress_tracker.epoch -
                            progress_tracker.last_improvement_epoch
                        )
                    )
                should_break = True
        return should_break

    def predict(
            self,
            dataset,
            batch_size,
            only_predictions=False,
            gpus=None,
            gpu_fraction=1,
            **kwargs
    ):
        if self.session is None:
            session = self.initialize_session(gpus, gpu_fraction)

            # load parameters
            if self.weights_save_path:
                self.restore(session, self.weights_save_path)
        else:
            session = self.session

        # predict
        predict_stats = self.batch_evaluation(
            session,
            dataset,
            batch_size,
            is_training=False,
            collect_predictions=True,
            only_predictions=only_predictions
        )

        return predict_stats

    def collect_activations(
            self,
            dataset,
            tensor_names,
            batch_size,
            gpus=None,
            gpu_fraction=1,
            **kwargs
    ):
        if self.session is None:
            session = self.initialize_session(gpus, gpu_fraction)

            # load parameters
            if self.weights_save_path:
                self.restore(session, self.weights_save_path)
        else:
            session = self.session

        # get operation names
        operation_names = set(
            [t.name for op in self.graph.get_operations() for t in op.values()]
        )
        for tensor_name in tensor_names:
            if tensor_name not in operation_names:
                raise ValueError(
                    'Tensor / operation {} not present in the '
                    'model graph'.format(tensor_name)
                )

        # collect tensors
        collected_tensors = self.batch_collect_activations(
            session,
            dataset,
            batch_size,
            tensor_names
        )

        return collected_tensors

    def collect_weights(
            self,
            tensor_names,
            gpus=None,
            gpu_fraction=1,
            **kwargs
    ):
        if self.session is None:
            session = self.initialize_session(gpus, gpu_fraction)

            # load parameters
            if self.weights_save_path:
                self.restore(session, self.weights_save_path)
        else:
            session = self.session

        operation_names = set(
            [t.name for op in self.graph.get_operations() for t in op.values()]
        )
        for tensor_name in tensor_names:
            if tensor_name not in operation_names:
                raise ValueError(
                    'Tensor / operation {} not present in the '
                    'model graph'.format(tensor_name)
                )

        # collect tensors
        collected_tensors = {
            tensor_name: session.run(self.graph.get_tensor_by_name(tensor_name))
            for tensor_name in tensor_names
        }

        return collected_tensors

    def save_weights(self, session, save_path):
        if is_on_master():
            self.weights_save_path = self.saver.save(session, save_path)
        self.weights_save_path = self.saver.save(session, save_path)

    def save_hyperparameters(self, hyperparameters, save_path):
        # removing pretrained embeddings paths from hyperparameters
        # because the weights are already saved in the model, no need to reload
        # from their path when loading the model next time

        local_hyperparamters = copy.deepcopy(hyperparameters)
        for feature in (local_hyperparamters['input_features'] +
                        local_hyperparamters['output_features']):
            if 'pretrained_embeddings' in feature:
                feature['pretrained_embeddings'] = None
        with open(save_path, 'w') as fp:
            json.dump(hyperparameters, fp, sort_keys=True, indent=4)

    def restore(self, session, weights_path):
        self.saver.restore(session, weights_path)

    @staticmethod
    def load(load_path, use_horovod=False):
        hyperparameter_file = os.path.join(
            load_path,
            MODEL_HYPERPARAMETERS_FILE_NAME
        )
        hyperparameters = load_json(hyperparameter_file)
        model = Model(use_horovod=use_horovod, **hyperparameters)
        model.weights_save_path = os.path.join(
            load_path,
            MODEL_WEIGHTS_FILE_NAME
        )
        return model

    def set_epochs_to_1_or_quit(self, signum, frame):
        if not self.received_sigint:
            self.epochs = 1
            self.received_sigint = True
            logging.critical(
                '\nReceived SIGINT, will finish this epoch and then conclude '
                'the training'
            )
            logging.critical(
                'Send another SIGINT to immediately interrupt the process'
            )
        else:
            logging.critical('\nReceived a second SIGINT, will now quit')
            sys.exit(1)

    def quit_training(self, signum, frame):
        logging.critical('Received SIGQUIT, will kill training')
        sys.exit(1)

    def resume_training(self, save_path, model_weights_path):
        if is_on_master():
            logging.info('Resuming training of model: {0}'.format(save_path))
        self.weights_save_path = model_weights_path
        progress_tracker = load_object(
            os.path.join(
                save_path,
                TRAINING_PROGRESS_FILE_NAME
            )
        )
        return progress_tracker

    def initialize_training_stats(self, output_features):
        train_stats = OrderedDict()
        vali_stats = OrderedDict()
        test_stats = OrderedDict()

        for output_feature in output_features:
            field_name = output_feature['name']

            train_stats[field_name] = OrderedDict()
            vali_stats[field_name] = OrderedDict()
            test_stats[field_name] = OrderedDict()
            output_config = output_type_registry[
                output_feature['type']].output_config

            for stat, config in output_config.items():
                if config['type'] == MEASURE:
                    train_stats[field_name][stat] = []
                    vali_stats[field_name][stat] = []
                    test_stats[field_name][stat] = []

        for stats in [train_stats, vali_stats, test_stats]:
            stats['combined'] = {
                LOSS: [],
                ACCURACY: []
            }

        return train_stats, vali_stats, test_stats

    def get_stat_names(self, output_features):
        stat_names = {}
        for output_feature in output_features:
            field_name = output_feature['name']
            output_config = output_type_registry[
                output_feature['type']].output_config

            for stat, config in output_config.items():
                if config['type'] == MEASURE:
                    stats = stat_names.get(field_name, [])
                    stats.append(stat)
                    stat_names[field_name] = stats
        stat_names['combined'] = [LOSS, ACCURACY]
        return stat_names

    def initialize_batcher(
            self,
            dataset,
            batch_size=128,
            bucketing_field=None,
            should_shuffle=True,
            ignore_last=False
    ):
        if self.horovod:
            batcher = DistributedBatcher(
                dataset,
                self.horovod.rank(),
                self.horovod,
                batch_size,
                should_shuffle=should_shuffle,
                ignore_last=ignore_last
            )
        elif bucketing_field is not None:
            input_features = self.hyperparameters['input_features']
            bucketing_feature = [
                feature for feature in input_features if
                feature['name'] == bucketing_field
            ]
            if not bucketing_feature:
                raise ValueError(
                    'Bucketing field {} not present in input features'.format(
                        bucketing_field
                    )
                )
            else:
                bucketing_feature = bucketing_feature[0]
            should_trim = bucketing_feature[
                              'encoder'] in dynamic_length_encoders
            if 'preprocessing' in bucketing_feature:
                trim_side = bucketing_feature['preprocessing']['padding']
            else:
                trim_side = self.hyperparameters['preprocessing'][
                    bucketing_feature['type']]['padding']

            batcher = BucketedBatcher(
                dataset,
                bucketing_field=bucketing_field,
                batch_size=batch_size,
                buckets=10,
                ignore_last=ignore_last,
                should_shuffle=should_shuffle,
                should_trim=should_trim,
                trim_side=trim_side
            )
        else:
            batcher = Batcher(
                dataset,
                batch_size,
                should_shuffle=should_shuffle,
                ignore_last=ignore_last
            )
        return batcher

    def resume_session(
            self,
            session,
            save_path,
            model_weights_path,
            model_weights_progress_path
    ):
        num_matching_files = 0
        pattern = re.compile(MODEL_WEIGHTS_PROGRESS_FILE_NAME)
        for file_path in os.listdir(save_path):
            if pattern.match(file_path):
                num_matching_files += 1
        if num_matching_files == 3:
            self.restore(session, model_weights_progress_path)
        else:
            self.restore(session, model_weights_path)

    def reduce_learning_rate(
            self,
            progress_tracker,
            reduce_learning_rate_on_plateau,
            reduce_learning_rate_on_plateau_patience,
            reduce_learning_rate_on_plateau_rate
    ):
        if (progress_tracker.last_improvement >=
                reduce_learning_rate_on_plateau_patience):
            if (progress_tracker.num_reductions_lr >=
                    reduce_learning_rate_on_plateau):
                if is_on_master():
                    logging.info('\n')
                    logging.info(
                        "It has been " +
                        str(progress_tracker.last_improvement) +
                        'epochs since last validation accuracy improvement '
                        'and the learning rate was already reduced ' +
                        str(progress_tracker.num_reductions_lr) +
                        'times, not reducing it anymore'
                    )
                    logging.info('\n')
            else:
                if is_on_master():
                    logging.info('\n')
                    logging.info(
                        'PLATEAU REACHED, reducing learning rate'
                        ' due to lack of validation improvement, it has been' +
                        str(progress_tracker.last_improvement) +
                        'epochs since last validation accuracy improvement'
                        'or since the learning rate was reduced'
                    )

                progress_tracker.learning_rate *= (
                    reduce_learning_rate_on_plateau_rate
                )
                progress_tracker.last_improvement_epoch = (
                    progress_tracker.epoch
                )
                progress_tracker.last_improvement = 0
                progress_tracker.num_reductions_lr += 1

    def increase_batch_size(
            self,
            progress_tracker,
            increase_batch_size_on_plateau_patience,
            increase_batch_size_on_plateau,
            increase_batch_size_on_plateau_max,
            increase_batch_size_on_plateau_rate
    ):
        if (progress_tracker.last_improvement >=
                increase_batch_size_on_plateau_patience):
            if (progress_tracker.num_increases_bs >=
                    increase_batch_size_on_plateau):
                if is_on_master():
                    logging.info('\n')
                    logging.info(
                        "It has been " +
                        str(progress_tracker.last_improvement) +
                        'epochs since last validation accuracy improvement '
                        'and the learning rate was already reduced ' +
                        str(progress_tracker.num_increases_bs) +
                        'times, not reducing it anymore'
                    )
                    logging.info('\n')

            elif (progress_tracker.batch_size ==
                  increase_batch_size_on_plateau_max):
                if is_on_master():
                    logging.info('\n')
                    logging.info(
                        "It has been" +
                        str(progress_tracker.last_improvement) +
                        'epochs since last validation accuracy improvement'
                        'and the batch size was already increased' +
                        str(progress_tracker.num_increases_bs) +
                        'times and currently is' +
                        str(progress_tracker.batch_size) +
                        ', the maximum allowed'
                    )
                    logging.info('\n')
            else:
                if is_on_master():
                    logging.info('\n')
                    logging.info(
                        'PLATEAU REACHED'
                        'increasing batch size due to lack of'
                        'validation improvement, it has been' +
                        str(progress_tracker.last_improvement) +
                        'epochs since last validation accuracy improvement'
                        'or since the batch size was increased'
                    )

                progress_tracker.batch_size = min(
                    (increase_batch_size_on_plateau_rate *
                     progress_tracker.batch_size),
                    increase_batch_size_on_plateau_max
                )
                progress_tracker.last_improvement_epoch = progress_tracker.epoch
                progress_tracker.last_improvement = 0
                progress_tracker.num_increases_bs += 1


class ProgressTracker:
    def __init__(
            self,
            epoch,
            batch_size,
            steps,
            last_improvement_epoch,
            best_valid_measure,
            learning_rate,
            num_reductions_lr,
            num_increases_bs,
            train_stats,
            vali_stats,
            test_stats
    ):
        self.batch_size = batch_size
        self.epoch = epoch
        self.steps = steps
        self.last_improvement_epoch = last_improvement_epoch
        self.last_improvement = 0
        self.learning_rate = learning_rate
        self.best_valid_measure = best_valid_measure
        self.num_reductions_lr = num_reductions_lr
        self.num_increases_bs = num_increases_bs
        self.train_stats = train_stats
        self.vali_stats = vali_stats
        self.test_stats = test_stats


def load_model_and_definition(model_dir, use_horovod=False):
    # Load model definition and weights
    model_definition = load_json(
        os.path.join(
            model_dir,
            MODEL_HYPERPARAMETERS_FILE_NAME
        )
    )
    model = Model.load(model_dir, use_horovod=use_horovod)
    return model, model_definition



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import csv
import os
import random
import string
import uuid

import numpy as np
import yaml
from skimage.io import imsave

from ludwig.utils.misc import get_from_registry

letters = string.ascii_letters


def generate_string(length):
    sequence = []
    for _ in range(length):
        sequence.append(random.choice(letters))
    return ''.join(sequence)


def build_vocab(size):
    vocab = []
    for _ in range(size):
        vocab.append(generate_string(random.randint(2, 10)))
    return vocab


def return_none(feature):
    return None


def assign_vocab(feature):
    feature['idx2str'] = build_vocab(feature['vocab_size'])


def build_feature_parameters(features):
    feature_parameters = {}
    for feature in features:
        fearure_builder_function = get_from_registry(
            feature['type'],
            parameters_builders_registry
        )

        feature_parameters[feature['name']] = fearure_builder_function(feature)
    return feature_parameters


parameters_builders_registry = {
    'category': assign_vocab,
    'text': assign_vocab,
    'numerical': return_none,
    'binary': return_none,
    'set': assign_vocab,
    'bag': assign_vocab,
    'sequence': assign_vocab,
    'timeseries': return_none,
    'image': return_none
}


def build_synthetic_dataset(dataset_size, features):
    build_feature_parameters(features)
    header = []
    for feature in features:
        header.append(feature['name'])

    yield header
    for _ in range(dataset_size):
        yield generate_datapoint(features)


def generate_datapoint(features):
    datapoint = []
    for feature in features:
        if ('cycle' in feature and feature['cycle'] == True and
                feature['type'] in cyclers_registry):
            cycler_function = cyclers_registry[feature['type']]
            feature_value = cycler_function(feature)
        else:
            generator_function = get_from_registry(
                feature['type'],
                generators_registry
            )
            feature_value = generator_function(feature)
        datapoint.append(feature_value)
    return datapoint


def generate_category(feature):
    return random.choice(feature['idx2str'])


def generate_text(feature):
    text = []
    for _ in range(random.randint(feature['max_len'] -
                                  int(feature['max_len'] * 0.2),
                                  feature['max_len'])):
        text.append(random.choice(feature['idx2str']))
    return ' '.join(text)


def generate_numerical(feature):
    return random.uniform(
        feature['min'] if 'min' in feature else 0,
        feature['max'] if 'max' in feature else 1
    )


def generate_binary(feature):
    p = feature['prob'] if 'prob' in feature else 0.5
    return np.random.choice([True, False], p=[p, 1 - p])


def generate_sequence(feature):
    length = feature['max_len']
    if 'min_len' in feature:
        length = random.randint(feature['min_len'], feature['max_len'])

    sequence = [random.choice(feature['idx2str']) for _ in range(length)]

    return ' '.join(sequence)


def generate_set(feature):
    elems = []
    for _ in range(random.randint(0, feature['max_len'])):
        elems.append(random.choice(feature['idx2str']))
    return ' '.join(list(set(elems)))


def generate_bag(feature):
    elems = []
    for _ in range(random.randint(0, feature['max_len'])):
        elems.append(random.choice(feature['idx2str']))
    return ' '.join(elems)


def generate_timeseries(feature):
    series = []
    for _ in range(feature['max_len']):
        series.append(
            str(
                random.uniform(
                    feature['min'] if 'min' in feature else 0,
                    feature['max'] if 'max' in feature else 1
                )
            )
        )
    return ' '.join(series)


def generate_image(feature):
    # Read num_channels, width, height
    num_channels = feature['num_channels']
    width = feature['width']
    height = feature['height']
    image_dest_folder = feature['destination_folder']

    if width <= 0 or height <= 0 or num_channels < 1:
        raise ValueError('Invalid arguments for generating images')

    # Create a Random Image
    if num_channels == 1:
        img = np.random.rand(width, height) * 255
    else:
        img = np.random.rand(width, height, num_channels) * 255.0

    # Generate a unique random filename
    image_filename = uuid.uuid4().hex[:10].upper() + '.jpg'

    # Save the image to disk either in a specified location/new folder
    try:
        if not os.path.exists(image_dest_folder):
            os.mkdir(image_dest_folder)

        image_dest_path = os.path.join(image_dest_folder, image_filename)
        imsave(image_dest_path, img.astype('uint8'))

    except IOError as e:
        raise IOError('Unable to create a folder for images/save image to disk.'
                      '{0}'.format(e))

    return image_dest_path


generators_registry = {
    'category': generate_category,
    'text': generate_sequence,
    'numerical': generate_numerical,
    'binary': generate_binary,
    'set': generate_set,
    'bag': generate_bag,
    'sequence': generate_sequence,
    'timeseries': generate_timeseries,
    'image': generate_image
}

category_cycle = 0


def cycle_category(feature):
    global category_cycle
    if category_cycle >= len(feature['idx2str']):
        category_cycle = 0
    category = feature['idx2str'][category_cycle]
    category_cycle += 1
    return category


binary_cycle = False


def cycle_binary(feature):
    global binary_cycle
    if binary_cycle:
        binary_cycle = False
        return True
    else:
        binary_cycle = True
        return False


cyclers_registry = {
    'category': cycle_category,
    'binary': cycle_binary
}


def write_csv(dataset, csv_file_path):
    with open(csv_file_path, 'w') as file:
        writer = csv.writer(file)
        for row in dataset:
            writer.writerow(row)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='This script generates a synthetic dataset.')
    parser.add_argument('csv_file_path', help='output csv file path')
    parser.add_argument(
        '-d',
        '--dataset_size',
        help='size of the dataset',
        type=int,
        default=100
    )
    parser.add_argument(
        '-f',
        '--features',
        default='[\
          {name: text_1, type: text, vocab_size: 20, max_len: 20}, \
          {name: text_2, type: text, vocab_size: 20, max_len: 20}, \
          {name: category_1, type: category, vocab_size: 10}, \
          {name: category_2, type: category, vocab_size: 15}, \
          {name: numerical_1, type: numerical}, \
          {name: numerical_2, type: numerical}, \
          {name: binary_1, type: binary}, \
          {name: binary_2, type: binary}, \
          {name: set_1, type: set, vocab_size: 20, max_len: 20}, \
          {name: set_2, type: set, vocab_size: 20, max_len: 20}, \
          {name: bag_1, type: bag, vocab_size: 20, max_len: 10}, \
          {name: bag_2, type: bag, vocab_size: 20, max_len: 10}, \
          {name: sequence_1, type: sequence, vocab_size: 20, max_len: 20}, \
          {name: sequence_2, type: sequence, vocab_size: 20, max_len: 20}, \
          {name: timeseries_1, type: timeseries, max_len: 20}, \
          {name: timeseries_2, type: timeseries, max_len: 20}, \
          ]',
        type=yaml.load, help='dataset features'
    )
    args = parser.parse_args()

    dataset = build_synthetic_dataset(args.dataset_size, args.features)
    write_csv(dataset, args.csv_file_path)



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from ludwig.models.modules.fully_connected_modules import FCStack


class Dense:
    def __init__(
            self,
            fc_layers=None,
            num_fc_layers=1,
            fc_size=256,
            norm=None,
            activation='relu',
            dropout=False,
            regularize=True,
            initializer=None,
            **kwargs
    ):
        self.fc_stack = FCStack(
            layers=fc_layers,
            num_layers=num_fc_layers,
            default_fc_size=fc_size,
            default_activation=activation,
            default_norm=norm,
            default_dropout=dropout,
            default_regularize=regularize,
            default_initializer=initializer
        )

    def __call__(
            self,
            input,
            input_size,
            regularizer,
            dropout_rate,
            is_training
    ):
        hidden = self.fc_stack(
            input,
            input_size,
            regularizer,
            dropout_rate,
            is_training=is_training
        )
        hidden_size = hidden.shape.as_list()[-1]

        return hidden, hidden_size



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import sys

from ludwig import collect
from ludwig import experiment
from ludwig import predict
from ludwig import train
from ludwig import visualize


class CLI(object):
    """CLI describes a command line interface for interacting with Ludwig, there
    are several different functions that can be performed. These functions are:
    - experiment - run an experiment using ludwig
    - predict - Given a list of $hat{y}$ values, compute $d(\hat{y}, y) under a
      specified metric
    - train - trains a model on the input file specified to it
    - visualize - Analysis of the results for the model on the dataset and
      presents a variety of plots to understand and evaluate the results
    - collect_weights - Collects the weights for a pretrained model as a tensor
      representation
    - collect_activations - For each datapoint, there exists a corresponding
      tensor representation which are collected through this method
    """

    def __init__(self):
        parser = argparse.ArgumentParser(
            description='ludwig cli runner',
            usage='''ludwig <command> [<args>]

Available sub-commands:
   experiment            Runs a full experiment training a model and testing it
   train                 Trains a model
   predict               Predicts using a pretrained model
   visualize             Visualizes experimental results
   collect_weights       Collects tensors containing a pretrained model weights
   collect_activations   Collects tensors for each datapoint using a pretrained model
''')
        parser.add_argument('command', help='Subcommand to run')
        # parse_args defaults to [1:] for args, but you need to
        # exclude the rest of the args too, or validation will fail
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            print('Unrecognized command')
            parser.print_help()
            exit(1)
        # use dispatch pattern to invoke method with same name
        getattr(self, args.command)()

    def experiment(self):
        experiment.cli(sys.argv[2:])

    def train(self):
        train.cli(sys.argv[2:])

    def predict(self):
        predict.cli(sys.argv[2:])

    def visualize(self):
        visualize.cli(sys.argv[2:])

    def collect_weights(self):
        collect.cli_collect_weights(sys.argv[2:])

    def collect_activations(self):
        collect.cli_collect_activations(sys.argv[2:])


def main():
    CLI()


if __name__ == '__main__':
    main()



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from ludwig.constants import *
from ludwig.features.feature_registries import base_type_registry
from ludwig.features.feature_registries import input_type_registry
from ludwig.features.feature_registries import output_type_registry
from ludwig.utils.misc import get_from_registry
from ludwig.utils.misc import merge_dict
from ludwig.utils.misc import set_default_value

default_random_seed = 42

default_preprocessing_force_split = False
default_preprocessing_split_probabilities = (0.7, 0.1, 0.2)
default_preprocessing_stratify = None

default_preprocessing_parameters = {
    'force_split': default_preprocessing_force_split,
    'split_probabilities': default_preprocessing_split_probabilities,
    'stratify': default_preprocessing_stratify
}
default_preprocessing_parameters.update({
    name: base_type.preprocessing_defaults for name, base_type in
base_type_registry.items()
})

default_combiner_type = 'concat'

default_training_params = {
    'optimizer': {'type': 'adam'},
    'epochs': 200,
    'regularizer': 'l2',
    'regularization_lambda': 0,
    'learning_rate': 0.001,
    'batch_size': 128,
    'dropout_rate': 0.0,
    'early_stop': 3,
    'reduce_learning_rate_on_plateau': 0,
    'reduce_learning_rate_on_plateau_patience': 5,
    'reduce_learning_rate_on_plateau_rate': 0.5,
    'increase_batch_size_on_plateau': 0,
    'increase_batch_size_on_plateau_patience': 5,
    'increase_batch_size_on_plateau_rate': 2,
    'increase_batch_size_on_plateau_max': 512,
    'decay': False,
    'decay_steps': 10000,
    'decay_rate': 0.96,
    'staircase': False,
    'gradient_clipping': None,
    'validation_field': 'combined',
    'validation_measure': LOSS,
    'bucketing_field': None,
    'learning_rate_warmup_epochs': 5
}

default_optimizer_params_registry = {
    'adam': {
        'beta1': 0.9,
        'beta2': 0.999,
        'epsilon': 1e-08
    },
    'adadelta': {
        'rho': 0.95,
        'epsilon': 1e-08
    },
    'adagrad': {
        'initial_accumulator_value': 0.1
    },
    'adagradda': {
        'initial_gradient_squared_accumulator_value': 0.1,
        'l1_regularization_strength': 0.0,
        'l2_regularization_strength': 0.0
    },
    'momentum': {
        'momentum': 0.1
    },
    'ftrl': {
        'learning_rate_power': -0.5,
        'initial_accumulator_value': 0.1,
        'l1_regularization_strength': 0.0,
        'l2_regularization_strength': 0.0
    },
    'proximalgd': {
        'l1_regularization_strength': 0.0,
        'l2_regularization_strength': 0.0
    },
    'proximaladagrad': {
        'initial_accumulator_value': 0.1,
        'l1_regularization_strength': 0.0,
        'l2_regularization_strength': 0.0
    },
    'rmsprop': {
        'decay': 0.9,
        'momentum': 0.0,
        'epsilon': 1e-10,
        'centered': False
    }
}


def get_default_optimizer_params(optimizer_type):
    if optimizer_type in default_optimizer_params_registry:
        return default_optimizer_params_registry[optimizer_type]
    else:
        raise ValueError('Incorrect optimizer type: ' + optimizer_type)


def merge_with_defaults(model_definition):
    # ===== Preprocessing =====
    model_definition['preprocessing'] = merge_dict(
        default_preprocessing_parameters,
        model_definition.get('preprocessing', {})
    )

    stratify = model_definition['preprocessing']['stratify']

    if stratify is not None:
        if stratify not in [x['name'] for x in
                            model_definition['output_features']]:
            raise ValueError('Stratify must be in output features')
        if ([x for x in model_definition['output_features'] if
             x['name'] == stratify][0]['type']
                not in [BINARY, CATEGORY]):
            raise ValueError('Stratify feature must be binary or category')
    # ===== Model =====
    set_default_value(model_definition, 'combiner',
                      {'type': default_combiner_type})

    # ===== Training =====
    set_default_value(model_definition, 'training', default_training_params)

    for param in default_training_params:
        set_default_value(model_definition['training'], param,
                          default_training_params[param])

    if len(model_definition['output_features']) == 0:
        raise ValueError('No output features are defined')

    set_default_value(
        model_definition['training'],
        'validation_measure',
        output_type_registry[model_definition['output_features'][0][
            'type']].default_validation_measure
    )

    # ===== Training Optimizer =====
    optimizer = model_definition['training']['optimizer']
    default_optimizer_params = get_default_optimizer_params(optimizer['type'])
    for param in default_optimizer_params:
        set_default_value(optimizer, param, default_optimizer_params[param])

    # ===== Input Features =====
    for input_feature in model_definition['input_features']:
        get_from_registry(input_feature['type'],
                          input_type_registry).populate_defaults(input_feature)

    # ===== Output features =====
    for output_feature in model_definition['output_features']:
        get_from_registry(output_feature['type'],
                          output_type_registry).populate_defaults(
            output_feature)

    return model_definition



#! /usr/bin/env python
# coding=utf-8
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import os

import h5py
import numpy as np
import tensorflow as tf

from skimage.io import imread

from ludwig.constants import *
from ludwig.features.base_feature import BaseFeature
from ludwig.features.base_feature import InputFeature
from ludwig.models.modules.image_encoders import ResNetEncoder
from ludwig.models.modules.image_encoders import Stacked2DCNN
from ludwig.utils.image_utils import resize_image
from ludwig.utils.misc import get_from_registry
from ludwig.utils.misc import set_default_value


class ImageBaseFeature(BaseFeature):
    def __init__(self, feature):
        super().__init__(feature)
        self.type = IMAGE

    preprocessing_defaults = {
        'missing_value_strategy': BACKFILL
    }

    @staticmethod
    def get_feature_meta(column, preprocessing_parameters):
        return {}

    @staticmethod
    def add_feature_data(
            feature,
            dataset_df,
            data,
            metadata,
            preprocessing_parameters
    ):
        csv_path = os.path.dirname(os.path.abspath(dataset_df.csv))
        an_image = imread(
            os.path.join(csv_path, dataset_df[feature['name']][0])
        )
        im_height = an_image.shape[0]
        im_width = an_image.shape[1]

        if an_image.ndim == 2:
            num_channels = 1
        else:
            num_channels = an_image.shape[2]

        num_images = len(dataset_df)

        if feature['should_resize']:
            im_height = feature[HEIGHT]
            im_width = feature[WIDTH]

        metadata[feature['name']] = {
            'height': im_height,
            'width': im_width,
            'num_channels': num_channels,
            'in_memory': feature['in_memory']
        }

        if feature['in_memory']:
            data[feature['name']] = np.empty(
                (num_images, im_height, im_width, num_channels),
                dtype=np.int8
            )
            for i in range(len(dataset_df)):
                filename = os.path.join(
                    csv_path,
                    dataset_df[feature['name']][i]
                )
                img = imread(filename)
                if img.ndim == 2:
                    img = img.reshape((img.shape[0], img.shape[1], 1))
                if feature['should_resize']:
                    img = resize_image(
                        img,
                        (im_height, im_width),
                        feature['resize_method']
                    )
                data[feature['name']][i, :, :, :] = img
        else:
            data_fp = dataset_df.csv.replace('csv', 'hdf5')
            mode = 'w'
            if os.path.isfile(data_fp):
                mode = 'r+'
            with h5py.File(data_fp, mode) as h5_file:
                image_dataset = h5_file.create_dataset(
                    feature['name'] + '_data',
                    (num_images, im_height, im_width, num_channels),
                    dtype=np.uint8
                )
                for i in range(len(dataset_df)):
                    filename = os.path.join(
                        csv_path,
                        dataset_df[feature['name']][i]
                    )
                    img = imread(filename)
                    if img.ndim == 2:
                        img = img.reshape((img.shape[0], img.shape[1], 1))
                    if feature['should_resize']:
                        img = resize_image(
                            img,
                            (im_height, im_width),
                            feature['resize_method'],
                        )

                    image_dataset[i, :im_height, :im_width, :] = img

            data[feature['name']] = np.arange(num_images)


class ImageInputFeature(ImageBaseFeature, InputFeature):
    def __init__(self, feature):
        super().__init__(feature)

        self.should_resize = False

        self.height = 0
        self.width = 0
        self.num_channels = 0

        self.in_memory = True
        self.data_hdf5_fp = ''

        self.encoder = 'stacked_cnn'

        encoder_parameters = self.overwrite_defaults(feature)

        self.encoder_obj = self.get_image_encoder(encoder_parameters)

    def get_image_encoder(self, encoder_parameters):
        return get_from_registry(
            self.encoder, image_encoder_registry)(
            **encoder_parameters
        )

    def _get_input_placeholder(self):
        # None dimension is for dealing with variable batch size
        return tf.placeholder(
            tf.float32,
            shape=[None, self.height, self.width, self.num_channels],
            name=self.name,
        )

    def build_input(
            self,
            regularizer,
            dropout_rate,
            is_training=False,
            **kwargs
    ):
        placeholder = self._get_input_placeholder()
        logging.debug('  targets_placeholder: {0}'.format(placeholder))

        feature_representation, feature_representation_size = self.encoder_obj(
            placeholder,
            regularizer,
            dropout_rate,
            is_training,
        )
        logging.debug(
            '  feature_representation: {0}'.format(feature_representation)
        )

        feature_representation = {
            'name': self.name,
            'type': self.type,
            'representation': feature_representation,
            'size': feature_representation_size,
            'placeholder': placeholder
        }
        return feature_representation

    @staticmethod
    def update_model_definition_with_metadata(
            input_feature,
            feature_metadata,
            *args,
            **kwargs
    ):
        for dim in ['height', 'width', 'num_channels']:
            input_feature[dim] = feature_metadata[dim]
        input_feature['data_hdf5_fp'] = (
            kwargs['model_definition']['data_hdf5_fp']
        )

    @staticmethod
    def populate_defaults(input_feature):
        set_default_value(input_feature, 'in_memory', True)

        if 'height' in input_feature or 'width' in input_feature:
            input_feature['should_resize'] = True
            try:
                input_feature[HEIGHT] = int(input_feature[HEIGHT])
                input_feature[WIDTH] = int(input_feature[WIDTH])
            except ValueError as e:
                raise ValueError(
                    'Image height and width must be set and have '
                    'positive integer values: ' + str(e)
                )
            if input_feature[HEIGHT] <= 0 or input_feature[WIDTH] <= 0:
                raise ValueError(
                    'Image height and width must be positive integers'
                )
            input_feature['should_resize'] = True
            set_default_value(input_feature, 'resize_method', 'crop_or_pad')
        else:
            input_feature['should_resize'] = False

        set_default_value(input_feature, 'tied_weights', None)


image_encoder_registry = {
    'stacked_cnn': Stacked2DCNN,
    'resnet': ResNetEncoder
}
